@@ -1,40 +1,33 @@
-/* -*- c-file-style: "linux"; -*-
-
-   Copyright (C) 1998-2001 by Andrew Tridgell <tridge@samba.org>
-   Copyright (C) 2000-2001 by Martin Pool <mbp@samba.org>
-
-   This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
-
-   This program is distributed in the hope that it will be useful,
-   but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-*/
-
 /*
-  Logging and utility functions.
-  tridge, May 1998
+ * Logging and utility functions.
+ *
+ * Copyright (C) 1998-2001 Andrew Tridgell <tridge@samba.org>
+ * Copyright (C) 2000-2001 Martin Pool <mbp@samba.org>
+ * Copyright (C) 2003-2008 Wayne Davison
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 3 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, visit the http://fsf.org website.
+ */
 
-  Mapping to human-readable messages added by Martin Pool
-  <mbp@samba.org>, Oct 2000.
-  */
 #include "rsync.h"
-#if defined HAVE_ICONV_OPEN && defined HAVE_ICONV_H
-#include <iconv.h>
-#endif
+#include "ifuncs.h"
 
 extern int verbose;
 extern int dry_run;
 extern int am_daemon;
 extern int am_server;
 extern int am_sender;
+extern int am_generator;
 extern int local_server;
 extern int quiet;
 extern int module_id;
@@ -42,23 +35,34 @@ extern int msg_fd_out;
 extern int allow_8bit_chars;
 extern int protocol_version;
 extern int preserve_times;
-extern int log_format_has_i;
-extern int log_format_has_o_or_i;
-extern int daemon_log_format_has_o_or_i;
+extern int uid_ndx;
+extern int gid_ndx;
+extern int stdout_format_has_i;
+extern int stdout_format_has_o_or_i;
+extern int logfile_format_has_i;
+extern int logfile_format_has_o_or_i;
+extern int receiver_symlink_times;
 extern mode_t orig_umask;
 extern char *auth_user;
-extern char *log_format;
-#if defined HAVE_ICONV_OPEN && defined HAVE_ICONV_H
+extern char *stdout_format;
+extern char *logfile_format;
+extern char *logfile_name;
+#ifdef ICONV_CONST
 extern iconv_t ic_chck;
 #endif
+#ifdef ICONV_OPTION
+extern iconv_t ic_send, ic_recv;
+#endif
+extern char curr_dir[];
+extern char *module_dir;
+extern unsigned int module_dirlen;
 
 static int log_initialised;
 static int logfile_was_closed;
-static char *logfname;
-static FILE *logfile;
+static FILE *logfile_fp;
 struct stats stats;
 
-int log_got_error = 0;
+int got_xfer_error = 0;
 
 struct {
         int code;
@@ -83,6 +87,7 @@ struct {
 	{ RERR_PARTIAL    , "some files could not be transferred" },
 	{ RERR_VANISHED   , "some files vanished before they could be transferred" },
 	{ RERR_TIMEOUT    , "timeout in data send/receive" },
+	{ RERR_CONTIMEOUT , "timeout waiting for daemon connection" },
 	{ RERR_CMD_FAILED , "remote shell failed" },
 	{ RERR_CMD_KILLED , "remote shell killed" },
 	{ RERR_CMD_RUN    , "remote command could not be run" },
@@ -91,7 +96,6 @@ struct {
 	{ 0, NULL }
 };
 
-
 /*
  * Map from rsync error code to name, or return NULL.
  */
@@ -105,14 +109,14 @@ static char const *rerr_name(int code)
 	return NULL;
 }
 
-static void logit(int priority, char *buf)
+static void logit(int priority, const char *buf)
 {
 	if (logfile_was_closed)
 		logfile_reopen();
-	if (logfile) {
-		fprintf(logfile,"%s [%d] %s",
+	if (logfile_fp) {
+		fprintf(logfile_fp, "%s [%d] %s",
 			timestring(time(NULL)), (int)getpid(), buf);
-		fflush(logfile);
+		fflush(logfile_fp);
 	} else {
 		syslog(priority, "%s", buf);
 	}
@@ -132,7 +136,7 @@ static void syslog_init()
 #endif
 
 #ifdef LOG_DAEMON
-	openlog("rsyncd", options, lp_syslog_facility());
+	openlog("rsyncd", options, lp_syslog_facility(module_id));
 #else
 	openlog("rsyncd", options);
 #endif
@@ -145,35 +149,49 @@ static void syslog_init()
 static void logfile_open(void)
 {
 	mode_t old_umask = umask(022 | orig_umask);
-	logfile = fopen(logfname, "a");
+	logfile_fp = fopen(logfile_name, "a");
 	umask(old_umask);
-	if (!logfile) {
+	if (!logfile_fp) {
 		int fopen_errno = errno;
 		/* Rsync falls back to using syslog on failure. */
 		syslog_init();
 		rsyserr(FERROR, fopen_errno,
-			"failed to open log-file %s", logfname);
+			"failed to open log-file %s", logfile_name);
 		rprintf(FINFO, "Ignoring \"log file\" setting.\n");
 	}
 }
 
-void log_init(void)
+void log_init(int restart)
 {
-	time_t t;
+	if (log_initialised) {
+		if (!restart)
+			return;
+		if (strcmp(logfile_name, lp_log_file(module_id)) != 0) {
+			if (logfile_fp) {
+				fclose(logfile_fp);
+				logfile_fp = NULL;
+			} else
+				closelog();
+			logfile_name = NULL;
+		} else if (*logfile_name)
+			return; /* unchanged, non-empty "log file" names */
+		else if (lp_syslog_facility(-1) != lp_syslog_facility(module_id))
+			closelog();
+		else
+			return; /* unchanged syslog settings */
+	} else
+		log_initialised = 1;
 
-	if (log_initialised)
-		return;
-	log_initialised = 1;
-
-	/* this looks pointless, but it is needed in order for the
+	/* This looks pointless, but it is needed in order for the
 	 * C library on some systems to fetch the timezone info
-	 * before the chroot */
-	t = time(NULL);
-	localtime(&t);
+	 * before the chroot. */
+	timestring(time(NULL));
 
-	/* optionally use a log file instead of syslog */
-	logfname = lp_log_file();
-	if (logfname && *logfname)
+	/* Optionally use a log file instead of syslog.  (Non-daemon
+	 * rsyncs will have already set logfile_name, as needed.) */
+	if (am_daemon && !logfile_name)
+		logfile_name = lp_log_file(module_id);
+	if (logfile_name && *logfile_name)
 		logfile_open();
 	else
 		syslog_init();
@@ -181,10 +199,10 @@ void log_init(void)
 
 void logfile_close(void)
 {
-	if (logfile) {
+	if (logfile_fp) {
 		logfile_was_closed = 1;
-		fclose(logfile);
-		logfile = NULL;
+		fclose(logfile_fp);
+		logfile_fp = NULL;
 	}
 }
 
@@ -202,11 +220,11 @@ static void filtered_fwrite(FILE *f, const char *buf, int len, int use_isprint)
 	for (s = buf; s < end; s++) {
 		if ((s < end - 4
 		  && *s == '\\' && s[1] == '#'
-		  && isdigit(*(uchar*)(s+2))
-		  && isdigit(*(uchar*)(s+3))
-		  && isdigit(*(uchar*)(s+4)))
+		  && isDigit(s + 2)
+		  && isDigit(s + 3)
+		  && isDigit(s + 4))
 		 || (*s != '\t'
-		  && ((use_isprint && !isprint(*(uchar*)s))
+		  && ((use_isprint && !isPrint(s))
 		   || *(uchar*)s < ' '))) {
 			if (s != buf && fwrite(buf, s - buf, 1, f) != 1)
 				exit_cleanup(RERR_MESSAGEIO);
@@ -219,52 +237,67 @@ static void filtered_fwrite(FILE *f, const char *buf, int len, int use_isprint)
 }
 
 /* this is the underlying (unformatted) rsync debugging function. Call
- * it with FINFO, FERROR or FLOG.  Note: recursion can happen with
- * certain fatal conditions. */
-void rwrite(enum logcode code, char *buf, int len)
+ * it with FINFO, FERROR_*, FWARNING, FLOG, or FCLIENT.  Note: recursion
+ * can happen with certain fatal conditions. */
+void rwrite(enum logcode code, const char *buf, int len, int is_utf8)
 {
 	int trailing_CR_or_NL;
 	FILE *f = NULL;
+#ifdef ICONV_OPTION
+	iconv_t ic = is_utf8 && ic_recv != (iconv_t)-1 ? ic_recv : ic_chck;
+#else
+#ifdef ICONV_CONST
+	iconv_t ic = ic_chck;
+#endif
+#endif
 
 	if (len < 0)
 		exit_cleanup(RERR_MESSAGEIO);
 
-	if (quiet && code == FINFO)
-		return;
-
 	if (am_server && msg_fd_out >= 0) {
+		assert(!is_utf8);
 		/* Pass the message to our sibling. */
-		send_msg((enum msgcode)code, buf, len);
+		send_msg((enum msgcode)code, buf, len, 0);
 		return;
 	}
 
-	if (code == FSOCKERR) /* This gets simplified for a non-sibling. */
+	if (code == FERROR_SOCKET) /* This gets simplified for a non-sibling. */
 		code = FERROR;
 
 	if (code == FCLIENT)
 		code = FINFO;
-	else if (am_daemon) {
+	else if (am_daemon || logfile_name) {
 		static int in_block;
 		char msg[2048];
-		int priority = code == FERROR ? LOG_WARNING : LOG_INFO;
+		int priority = code == FINFO || code == FLOG ? LOG_INFO :  LOG_WARNING;
 
 		if (in_block)
 			return;
 		in_block = 1;
 		if (!log_initialised)
-			log_init();
+			log_init(0);
 		strlcpy(msg, buf, MIN((int)sizeof msg, len + 1));
 		logit(priority, msg);
 		in_block = 0;
 
-		if (code == FLOG || !am_server)
+		if (code == FLOG || (am_daemon && !am_server))
 			return;
 	} else if (code == FLOG)
 		return;
 
+	if (quiet && code == FINFO)
+		return;
+
 	if (am_server) {
+		enum msgcode msg = (enum msgcode)code;
+		if (protocol_version < 30) {
+			if (msg == MSG_ERROR)
+				msg = MSG_ERROR_XFER;
+			else if (msg == MSG_WARNING)
+				msg = MSG_INFO;
+		}
 		/* Pass the message to the non-server side. */
-		if (io_multiplex_write((enum msgcode)code, buf, len))
+		if (send_msg(msg, buf, len, !is_utf8))
 			return;
 		if (am_daemon) {
 			/* TODO: can we send the error to the user somehow? */
@@ -273,20 +306,14 @@ void rwrite(enum logcode code, char *buf, int len)
 	}
 
 	switch (code) {
+	case FERROR_XFER:
+		got_xfer_error = 1;
+		/* FALL THROUGH */
 	case FERROR:
-		log_got_error = 1;
+	case FWARNING:
 		f = stderr;
-		goto pre_scan;
-	case FINFO:
-		f = am_server ? stderr : stdout;
-	pre_scan:
-		while (len > 1 && *buf == '\n') {
-			fputc(*buf, f);
-			buf++;
-			len--;
-		}
 		break;
-	case FNAME:
+	case FINFO:
 		f = am_server ? stderr : stdout;
 		break;
 	default:
@@ -296,27 +323,27 @@ void rwrite(enum logcode code, char *buf, int len)
 	trailing_CR_or_NL = len && (buf[len-1] == '\n' || buf[len-1] == '\r')
 			  ? buf[--len] : 0;
 
-#if defined HAVE_ICONV_OPEN && defined HAVE_ICONV_H
-	if (ic_chck != (iconv_t)-1) {
+#ifdef ICONV_CONST
+	if (ic != (iconv_t)-1) {
+		xbuf outbuf, inbuf;
 		char convbuf[1024];
-		char *in_buf = buf, *out_buf = convbuf;
-		size_t in_cnt = len, out_cnt = sizeof convbuf - 1;
+		int ierrno;
 
-		iconv(ic_chck, NULL, 0, NULL, 0);
-		while (iconv(ic_chck, &in_buf,&in_cnt,
-				 &out_buf,&out_cnt) == (size_t)-1) {
-			if (out_buf != convbuf) {
-				filtered_fwrite(f, convbuf, out_buf - convbuf, 0);
-				out_buf = convbuf;
-				out_cnt = sizeof convbuf - 1;
+		INIT_CONST_XBUF(outbuf, convbuf);
+		INIT_XBUF(inbuf, (char*)buf, len, -1);
+
+		while (inbuf.len) {
+			iconvbufs(ic, &inbuf, &outbuf, 0);
+			ierrno = errno;
+			if (outbuf.len) {
+				filtered_fwrite(f, convbuf, outbuf.len, 0);
+				outbuf.len = 0;
 			}
-			if (errno == E2BIG)
+			if (!ierrno || ierrno == E2BIG)
 				continue;
-			fprintf(f, "\\#%03o", *(uchar*)in_buf++);
-			in_cnt--;
+			fprintf(f, "\\#%03o", CVAL(inbuf.buf, inbuf.pos++));
+			inbuf.len--;
 		}
-		if (out_buf != convbuf)
-			filtered_fwrite(f, convbuf, out_buf - convbuf, 0);
 	} else
 #endif
 		filtered_fwrite(f, buf, len, !allow_8bit_chars);
@@ -327,8 +354,8 @@ void rwrite(enum logcode code, char *buf, int len)
 	}
 }
 
-/* This is the rsync debugging function. Call it with FINFO, FERROR or
- * FLOG. */
+/* This is the rsync debugging function. Call it with FINFO, FERROR_*,
+ * FWARNING, FLOG, or FCLIENT. */
 void rprintf(enum logcode code, const char *format, ...)
 {
 	va_list ap;
@@ -365,7 +392,7 @@ void rprintf(enum logcode code, const char *format, ...)
 		}
 	}
 
-	rwrite(code, buf, len);
+	rwrite(code, buf, len, 0);
 }
 
 /* This is like rprintf, but it also tries to print some
@@ -382,7 +409,7 @@ void rsyserr(enum logcode code, int errcode, const char *format, ...)
 	char buf[BIGPATHBUFLEN];
 	size_t len;
 
-	strcpy(buf, RSYNC_NAME ": ");
+	strlcpy(buf, RSYNC_NAME ": ", sizeof buf);
 	len = (sizeof RSYNC_NAME ": ") - 1;
 
 	va_start(ap, format);
@@ -396,44 +423,33 @@ void rsyserr(enum logcode code, int errcode, const char *format, ...)
 	if (len >= sizeof buf)
 		exit_cleanup(RERR_MESSAGEIO);
 
-	rwrite(code, buf, len);
+	rwrite(code, buf, len, 0);
 }
 
 void rflush(enum logcode code)
 {
 	FILE *f = NULL;
 
-	if (am_daemon) {
+	if (am_daemon || code == FLOG)
 		return;
-	}
 
-	if (code == FLOG) {
-		return;
-	}
-
-	if (code == FERROR) {
+	if (code == FINFO && !am_server)
+		f = stdout;
+	else
 		f = stderr;
-	}
 
-	if (code == FINFO) {
-		if (am_server)
-			f = stderr;
-		else
-			f = stdout;
-	}
-
-	if (!f) exit_cleanup(RERR_MESSAGEIO);
 	fflush(f);
 }
 
-/* a generic logging routine for send/recv, with parameter
- * substitiution */
-static void log_formatted(enum logcode code, char *format, char *op,
-			  struct file_struct *file, struct stats *initial_stats,
-			  int iflags, char *hlink)
+/* A generic logging routine for send/recv, with parameter substitiution. */
+static void log_formatted(enum logcode code, const char *format, const char *op,
+			  struct file_struct *file, const char *fname,
+			  struct stats *initial_stats, int iflags,
+			  const char *hlink)
 {
 	char buf[MAXPATHLEN+1024], buf2[MAXPATHLEN], fmt[32];
-	char *p, *s, *n;
+	char *p, *s, *c;
+	const char *n;
 	size_t len, total;
 	int64 b;
 
@@ -452,14 +468,14 @@ static void log_formatted(enum logcode code, char *format, char *op,
 
 	for (p = buf; (p = strchr(p, '%')) != NULL; ) {
 		s = p++;
-		n = fmt + 1;
+		c = fmt + 1;
 		if (*p == '-')
-			*n++ = *p++;
-		while (isdigit(*(uchar*)p) && n - fmt < (int)(sizeof fmt) - 8)
-			*n++ = *p++;
+			*c++ = *p++;
+		while (isDigit(p) && c - fmt < (int)(sizeof fmt) - 8)
+			*c++ = *p++;
 		if (!*p)
 			break;
-		*n = '\0';
+		*c = '\0';
 		n = NULL;
 
 		switch (*p) {
@@ -474,22 +490,22 @@ static void log_formatted(enum logcode code, char *format, char *op,
 		case 'l':
 			strlcat(fmt, ".0f", sizeof fmt);
 			snprintf(buf2, sizeof buf2, fmt,
-				 (double)file->length);
+				 (double)F_LENGTH(file));
 			n = buf2;
 			break;
 		case 'U':
-			strlcat(fmt, "ld", sizeof fmt);
+			strlcat(fmt, "u", sizeof fmt);
 			snprintf(buf2, sizeof buf2, fmt,
-				 (long)file->uid);
+				 uid_ndx ? F_OWNER(file) : 0);
 			n = buf2;
 			break;
 		case 'G':
-			if (file->gid == GID_NONE)
+			if (!gid_ndx || file->flags & FLAG_SKIP_GROUP)
 				n = "DEFAULT";
 			else {
-				strlcat(fmt, "ld", sizeof fmt);
+				strlcat(fmt, "u", sizeof fmt);
 				snprintf(buf2, sizeof buf2, fmt,
-					 (long)file->gid);
+					 F_GROUP(file));
 				n = buf2;
 			}
 			break;
@@ -500,52 +516,71 @@ static void log_formatted(enum logcode code, char *format, char *op,
 			n = buf2;
 			break;
 		case 'M':
-			n = timestring(file->modtime);
-			{
-				char *cp = n;
-				while ((cp = strchr(cp, ' ')) != NULL)
-					*cp = '-';
-			}
+			n = c = timestring(file->modtime);
+			while ((c = strchr(p, ' ')) != NULL)
+				*c = '-';
 			break;
 		case 'B':
-			n = buf2 + MAXPATHLEN - PERMSTRING_SIZE;
-			permstring(n - 1, file->mode); /* skip the type char */
+			c = buf2 + MAXPATHLEN - PERMSTRING_SIZE - 1;
+			permstring(c, file->mode);
+			n = c + 1; /* skip the type char */
 			break;
 		case 'o':
 			n = op;
 			break;
 		case 'f':
-			n = f_name(file, NULL);
-			if (am_sender && file->dir.root) {
-				pathjoin(buf2, sizeof buf2,
-					 file->dir.root, n);
-				clean_fname(buf2, 0);
-				if (fmt[1])
-					strlcpy(n, buf2, MAXPATHLEN);
-				else
-					n = buf2;
+			if (fname) {
+				c = f_name_buf();
+				strlcpy(c, fname, MAXPATHLEN);
 			} else
-				clean_fname(n, 0);
+				c = f_name(file, NULL);
+			if (am_sender && F_PATHNAME(file)) {
+				pathjoin(buf2, sizeof buf2,
+					 F_PATHNAME(file), c);
+				clean_fname(buf2, 0);
+				if (fmt[1]) {
+					strlcpy(c, buf2, MAXPATHLEN);
+					n = c;
+				} else
+					n = buf2;
+			} else if (am_daemon && *c != '/') {
+				pathjoin(buf2, sizeof buf2,
+					 curr_dir + module_dirlen, c);
+				clean_fname(buf2, 0);
+				if (fmt[1]) {
+					strlcpy(c, buf2, MAXPATHLEN);
+					n = c;
+				} else
+					n = buf2;
+			} else {
+				clean_fname(c, 0);
+				n = c;
+			}
 			if (*n == '/')
 				n++;
 			break;
 		case 'n':
-			n = f_name(file, NULL);
+			if (fname) {
+				c = f_name_buf();
+				strlcpy(c, fname, MAXPATHLEN);
+			} else
+				c = f_name(file, NULL);
 			if (S_ISDIR(file->mode))
-				strlcat(n, "/", MAXPATHLEN);
+				strlcat(c, "/", MAXPATHLEN);
+			n = c;
 			break;
 		case 'L':
 			if (hlink && *hlink) {
 				n = hlink;
-				strcpy(buf2, " => ");
-			} else if (S_ISLNK(file->mode) && file->u.link) {
-				n = file->u.link;
-				strcpy(buf2, " -> ");
+				strlcpy(buf2, " => ", sizeof buf2);
+			} else if (S_ISLNK(file->mode) && !fname) {
+				n = F_SYMLINK(file);
+				strlcpy(buf2, " -> ", sizeof buf2);
 			} else {
 				n = "";
 				if (!fmt[1])
 					break;
-				strcpy(buf2, "    ");
+				strlcpy(buf2, "    ", sizeof buf2);
 			}
 			strlcat(fmt, "s", sizeof fmt);
 			snprintf(buf2 + 4, sizeof buf2 - 4, fmt, n);
@@ -558,7 +593,7 @@ static void log_formatted(enum logcode code, char *format, char *op,
 			n = timestring(time(NULL));
 			break;
 		case 'P':
-			n = lp_path(module_id);
+			n = module_dir;
 			break;
 		case 'u':
 			n = auth_user;
@@ -589,43 +624,45 @@ static void log_formatted(enum logcode code, char *format, char *op,
 			break;
 		case 'i':
 			if (iflags & ITEM_DELETED) {
-				n = "*deleting";
+				n = "*deleting  ";
 				break;
 			}
-			n = buf2 + MAXPATHLEN - 32;
-			n[0] = iflags & ITEM_LOCAL_CHANGE
+			n  = c = buf2 + MAXPATHLEN - 32;
+			c[0] = iflags & ITEM_LOCAL_CHANGE
 			      ? iflags & ITEM_XNAME_FOLLOWS ? 'h' : 'c'
 			     : !(iflags & ITEM_TRANSFER) ? '.'
 			     : !local_server && *op == 's' ? '<' : '>';
-			n[1] = S_ISDIR(file->mode) ? 'd'
+			c[1] = S_ISDIR(file->mode) ? 'd'
 			     : IS_SPECIAL(file->mode) ? 'S'
 			     : IS_DEVICE(file->mode) ? 'D'
 			     : S_ISLNK(file->mode) ? 'L' : 'f';
-			n[2] = !(iflags & ITEM_REPORT_CHECKSUM) ? '.' : 'c';
-			n[3] = !(iflags & ITEM_REPORT_SIZE) ? '.' : 's';
-			n[4] = !(iflags & ITEM_REPORT_TIME) ? '.'
-			     : !preserve_times || S_ISLNK(file->mode) ? 'T' : 't';
-			n[5] = !(iflags & ITEM_REPORT_PERMS) ? '.' : 'p';
-			n[6] = !(iflags & ITEM_REPORT_OWNER) ? '.' : 'o';
-			n[7] = !(iflags & ITEM_REPORT_GROUP) ? '.' : 'g';
-			n[8] = '.';
-			n[9] = '\0';
+			c[2] = !(iflags & ITEM_REPORT_CHECKSUM) ? '.' : 'c';
+			c[3] = !(iflags & ITEM_REPORT_SIZE) ? '.' : 's';
+			c[4] = !(iflags & ITEM_REPORT_TIME) ? '.'
+			     : !preserve_times || (!receiver_symlink_times && S_ISLNK(file->mode))
+			     ? 'T' : 't';
+			c[5] = !(iflags & ITEM_REPORT_PERMS) ? '.' : 'p';
+			c[6] = !(iflags & ITEM_REPORT_OWNER) ? '.' : 'o';
+			c[7] = !(iflags & ITEM_REPORT_GROUP) ? '.' : 'g';
+			c[8] = !(iflags & ITEM_REPORT_ATIME) ? '.' : 'u';
+			c[9] = !(iflags & ITEM_REPORT_ACL) ? '.' : 'a';
+			c[10] = !(iflags & ITEM_REPORT_XATTR) ? '.' : 'x';
+			c[11] = '\0';
 
 			if (iflags & (ITEM_IS_NEW|ITEM_MISSING_DATA)) {
 				char ch = iflags & ITEM_IS_NEW ? '+' : '?';
 				int i;
-				for (i = 2; n[i]; i++)
-					n[i] = ch;
-			} else if (n[0] == '.' || n[0] == 'h'
-			        || (n[0] == 'c' && n[1] == 'f')) {
+				for (i = 2; c[i]; i++)
+					c[i] = ch;
+			} else if (c[0] == '.' || c[0] == 'h' || c[0] == 'c') {
 				int i;
-				for (i = 2; n[i]; i++) {
-					if (n[i] != '.')
+				for (i = 2; c[i]; i++) {
+					if (c[i] != '.')
 						break;
 				}
-				if (!n[i]) {
-					for (i = 2; n[i]; i++)
-						n[i] = ' ';
+				if (!c[i]) {
+					for (i = 2; c[i]; i++)
+						c[i] = ' ';
 				}
 			}
 			break;
@@ -664,7 +701,7 @@ static void log_formatted(enum logcode code, char *format, char *op,
 		p = s + len;
 	}
 
-	rwrite(code, buf, total);
+	rwrite(code, buf, total, 0);
 }
 
 /* Return 1 if the format escape is in the log-format string (e.g. look for
@@ -679,7 +716,7 @@ int log_format_has(const char *format, char esc)
 	for (p = format; (p = strchr(p, '%')) != NULL; ) {
 		if (*++p == '-')
 			p++;
-		while (isdigit(*(uchar*)p))
+		while (isDigit(p))
 			p++;
 		if (!*p)
 			break;
@@ -689,62 +726,70 @@ int log_format_has(const char *format, char esc)
 	return 0;
 }
 
-/* log the transfer of a file */
-void log_item(struct file_struct *file, struct stats *initial_stats,
-	      int iflags, char *hlink)
+/* Log the transfer of a file.  If the code is FCLIENT, the output just goes
+ * to stdout.  If it is FLOG, it just goes to the log file.  Otherwise we
+ * output to both. */
+void log_item(enum logcode code, struct file_struct *file,
+	      struct stats *initial_stats, int iflags, const char *hlink)
 {
-	char *s_or_r = am_sender ? "send" : "recv";
+	const char *s_or_r = am_sender ? "send" : "recv";
 
-	if (lp_transfer_logging(module_id)) {
-		log_formatted(FLOG, lp_log_format(module_id), s_or_r,
-			      file, initial_stats, iflags, hlink);
-	} else if (log_format && !am_server) {
-		log_formatted(FNAME, log_format, s_or_r,
-			      file, initial_stats, iflags, hlink);
+	if (code != FLOG && stdout_format && !am_server) {
+		log_formatted(FCLIENT, stdout_format, s_or_r,
+			      file, NULL, initial_stats, iflags, hlink);
+	}
+	if (code != FCLIENT && logfile_format && *logfile_format) {
+		log_formatted(FLOG, logfile_format, s_or_r,
+			      file, NULL, initial_stats, iflags, hlink);
 	}
 }
 
 void maybe_log_item(struct file_struct *file, int iflags, int itemizing,
-		    char *buf)
+		    const char *buf)
 {
 	int significant_flags = iflags & SIGNIFICANT_ITEM_FLAGS;
 	int see_item = itemizing && (significant_flags || *buf
-		|| log_format_has_i > 1 || (verbose > 1 && log_format_has_i));
+		|| stdout_format_has_i > 1 || (verbose > 1 && stdout_format_has_i));
 	int local_change = iflags & ITEM_LOCAL_CHANGE && significant_flags;
 	if (am_server) {
-		if (am_daemon && !dry_run && see_item)
-			log_item(file, &stats, iflags, buf);
+		if (logfile_name && !dry_run && see_item
+		 && (significant_flags || logfile_format_has_i))
+			log_item(FLOG, file, &stats, iflags, buf);
 	} else if (see_item || local_change || *buf
-	    || (S_ISDIR(file->mode) && significant_flags))
-		log_item(file, &stats, iflags, buf);
+	    || (S_ISDIR(file->mode) && significant_flags)) {
+		enum logcode code = significant_flags || logfile_format_has_i ? FINFO : FCLIENT;
+		log_item(code, file, &stats, iflags, buf);
+	}
 }
 
-void log_delete(char *fname, int mode)
+void log_delete(const char *fname, int mode)
 {
-	static struct file_struct file;
+	static struct {
+		union file_extras ex[4]; /* just in case... */
+		struct file_struct file;
+	} x;
 	int len = strlen(fname);
-	char *fmt;
+	const char *fmt;
 
-	file.mode = mode;
-	file.basename = fname;
+	x.file.mode = mode;
 
-	if (!verbose && !log_format)
+	if (!verbose && !stdout_format)
 		;
 	else if (am_server && protocol_version >= 29 && len < MAXPATHLEN) {
 		if (S_ISDIR(mode))
 			len++; /* directories include trailing null */
-		send_msg(MSG_DELETED, fname, len);
+		send_msg(MSG_DELETED, fname, len, am_generator);
 	} else {
-		fmt = log_format_has_o_or_i ? log_format : "deleting %n";
-		log_formatted(FCLIENT, fmt, "del.", &file, &stats,
+		fmt = stdout_format_has_o_or_i ? stdout_format : "deleting %n";
+		log_formatted(FCLIENT, fmt, "del.", &x.file, fname, &stats,
 			      ITEM_DELETED, NULL);
 	}
 
-	if (!am_daemon || dry_run || !lp_transfer_logging(module_id))
+	if (!logfile_name || dry_run || !logfile_format)
 		return;
 
-	fmt = daemon_log_format_has_o_or_i ? lp_log_format(module_id) : "deleting %n";
-	log_formatted(FLOG, fmt, "del.", &file, &stats, ITEM_DELETED, NULL);
+	fmt = logfile_format_has_o_or_i ? logfile_format : "deleting %n";
+	log_formatted(FLOG, fmt, "del.", &x.file, fname, &stats, ITEM_DELETED, NULL);
 }
 
 /*
@@ -760,7 +805,7 @@ void log_exit(int code, const char *file, int line)
 			(double)stats.total_written,
 			(double)stats.total_read,
 			(double)stats.total_size);
-	} else {
+	} else if (am_server != 2) {
 		const char *name;
 
 		name = rerr_name(code);
@@ -769,7 +814,7 @@ void log_exit(int code, const char *file, int line)
 
 		/* VANISHED is not an error, only a warning */
 		if (code == RERR_VANISHED) {
-			rprintf(FINFO, "rsync warning: %s (code %d) at %s(%d) [%s=%s]\n", 
+			rprintf(FWARNING, "rsync warning: %s (code %d) at %s(%d) [%s=%s]\n",
 				name, code, file, line, who_am_i(), RSYNC_VERSION);
 		} else {
 			rprintf(FERROR, "rsync error: %s (code %d) at %s(%d) [%s=%s]\n",
