@@ -4,11 +4,11 @@
  * Copyright (C) 1996-2001 Andrew Tridgell <tridge@samba.org>
  * Copyright (C) 1996 Paul Mackerras
  * Copyright (C) 2001, 2002 Martin Pool <mbp@samba.org>
- * Copyright (C) 2003, 2004, 2005, 2006 Wayne Davison
+ * Copyright (C) 2003-2008 Wayne Davison
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -17,11 +17,12 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
+ * with this program; if not, visit the http://fsf.org website.
  */
 
 #include "rsync.h"
+#include "ifuncs.h"
+#include "io.h"
 #if defined CONFIG_LOCALE && defined HAVE_LOCALE_H
 #include <locale.h>
 #endif
@@ -34,20 +35,24 @@ extern int am_server;
 extern int am_sender;
 extern int am_generator;
 extern int am_daemon;
+extern int inc_recurse;
 extern int blocking_io;
 extern int remove_source_files;
-extern int daemon_over_rsh;
 extern int need_messages_from_generator;
 extern int kluge_around_eof;
 extern int do_stats;
-extern int log_got_error;
+extern int got_xfer_error;
 extern int module_id;
 extern int copy_links;
 extern int copy_dirlinks;
+extern int copy_unsafe_links;
 extern int keep_dirlinks;
 extern int preserve_hard_links;
 extern int protocol_version;
+extern int file_total;
 extern int recurse;
+extern int xfer_dirs;
+extern int protect_args;
 extern int relative_paths;
 extern int sanitize_paths;
 extern int curr_dir_depth;
@@ -58,23 +63,31 @@ extern int whole_file;
 extern int read_batch;
 extern int write_batch;
 extern int batch_fd;
-extern int batch_gen_fd;
 extern int filesfrom_fd;
+extern int connect_timeout;
 extern pid_t cleanup_child_pid;
+extern unsigned int module_dirlen;
 extern struct stats stats;
 extern char *filesfrom_host;
 extern char *partial_dir;
+extern char *dest_option;
 extern char *basis_dir[];
 extern char *rsync_path;
 extern char *shell_cmd;
 extern char *batch_name;
+extern char *password_file;
 extern char curr_dir[MAXPATHLEN];
+extern struct file_list *first_flist;
 extern struct filter_list_struct server_filter_list;
+#ifdef ICONV_OPTION
+extern iconv_t ic_send;
+#endif
 
+uid_t our_uid;
 int local_server = 0;
-int new_root_dir = 0;
+int daemon_over_rsh = 0;
 mode_t orig_umask = 0;
-struct file_list *the_file_list;
+int batch_gen_fd = -1;
 
 /* There's probably never more than at most 2 outstanding child processes,
  * but set it higher, just in case. */
@@ -103,7 +116,11 @@ static void show_malloc_stats(void);
  * remember_children(), we succeed instead of returning an error. */
 pid_t wait_process(pid_t pid, int *status_ptr, int flags)
 {
-	pid_t waited_pid = waitpid(pid, status_ptr, flags);
+	pid_t waited_pid;
+	
+	do {
+		waited_pid = waitpid(pid, status_ptr, flags);
+	} while (waited_pid == -1 && errno == EINTR);
 
 	if (waited_pid == -1 && errno == ECHILD) {
 		/* Status of requested child no longer available:  check to
@@ -185,12 +202,12 @@ static void handle_stats(int f)
 
 	if (am_server) {
 		if (am_sender) {
-			write_longint(f, total_read);
-			write_longint(f, total_written);
-			write_longint(f, stats.total_size);
+			write_varlong30(f, total_read, 3);
+			write_varlong30(f, total_written, 3);
+			write_varlong30(f, stats.total_size, 3);
 			if (protocol_version >= 29) {
-				write_longint(f, stats.flist_buildtime);
-				write_longint(f, stats.flist_xfertime);
+				write_varlong30(f, stats.flist_buildtime, 3);
+				write_varlong30(f, stats.flist_xfertime, 3);
 			}
 		}
 		return;
@@ -203,22 +220,22 @@ static void handle_stats(int f)
 	else if (!am_sender) {
 		/* Read the first two in opposite order because the meaning of
 		 * read/write swaps when switching from sender to receiver. */
-		total_written = read_longint(f);
-		total_read = read_longint(f);
-		stats.total_size = read_longint(f);
+		total_written = read_varlong30(f, 3);
+		total_read = read_varlong30(f, 3);
+		stats.total_size = read_varlong30(f, 3);
 		if (protocol_version >= 29) {
-			stats.flist_buildtime = read_longint(f);
-			stats.flist_xfertime = read_longint(f);
+			stats.flist_buildtime = read_varlong30(f, 3);
+			stats.flist_xfertime = read_varlong30(f, 3);
 		}
 	} else if (write_batch) {
 		/* The --read-batch process is going to be a client
 		 * receiver, so we need to give it the stats. */
-		write_longint(batch_fd, total_read);
-		write_longint(batch_fd, total_written);
-		write_longint(batch_fd, stats.total_size);
+		write_varlong30(batch_fd, total_read, 3);
+		write_varlong30(batch_fd, total_written, 3);
+		write_varlong30(batch_fd, stats.total_size, 3);
 		if (protocol_version >= 29) {
-			write_longint(batch_fd, stats.flist_buildtime);
-			write_longint(batch_fd, stats.flist_xfertime);
+			write_varlong30(batch_fd, stats.flist_buildtime, 3);
+			write_varlong30(batch_fd, stats.flist_xfertime, 3);
 		}
 	}
 }
@@ -238,7 +255,8 @@ static void output_summary(void)
 			human_num(stats.literal_data));
 		rprintf(FINFO,"Matched data: %s bytes\n",
 			human_num(stats.matched_data));
-		rprintf(FINFO,"File list size: %d\n", stats.flist_size);
+		rprintf(FINFO,"File list size: %s\n",
+			human_num(stats.flist_size));
 		if (stats.flist_buildtime) {
 			rprintf(FINFO,
 				"File list generation time: %.3f seconds\n",
@@ -259,9 +277,10 @@ static void output_summary(void)
 			"sent %s bytes  received %s bytes  %s bytes/sec\n",
 			human_num(total_written), human_num(total_read),
 			human_dnum((total_written + total_read)/(0.5 + (endtime - starttime)), 2));
-		rprintf(FINFO, "total size is %s  speedup is %.2f\n",
+		rprintf(FINFO, "total size is %s  speedup is %.2f%s\n",
 			human_num(stats.total_size),
-			(double)stats.total_size / (total_written+total_read));
+			(double)stats.total_size / (total_written+total_read),
+			write_batch < 0 ? " (BATCH ONLY)" : dry_run ? " (DRY RUN)" : "");
 	}
 
 	fflush(stdout);
@@ -310,13 +329,12 @@ static void show_malloc_stats(void)
 
 
 /* Start the remote shell.   cmd may be NULL to use the default. */
-static pid_t do_cmd(char *cmd, char *machine, char *user, char *path,
-		    int *f_in, int *f_out)
+static pid_t do_cmd(char *cmd, char *machine, char *user, char **remote_argv, int remote_argc,
+		    int *f_in_p, int *f_out_p)
 {
 	int i, argc = 0;
 	char *args[MAX_ARGS];
 	pid_t ret;
-	char *dir = NULL;
 	int dash_l_set = 0;
 
 	if (!read_batch && !local_server) {
@@ -334,10 +352,8 @@ static pid_t do_cmd(char *cmd, char *machine, char *user, char *path,
 			if (*f == ' ')
 				continue;
 			/* Comparison leaves rooms for server_options(). */
-			if (argc >= MAX_ARGS - MAX_SERVER_ARGS) {
-				rprintf(FERROR, "internal: args[] overflowed in do_cmd()\n");
-				exit_cleanup(RERR_SYNTAX);
-			}
+			if (argc >= MAX_ARGS - MAX_SERVER_ARGS)
+				goto arg_overflow;
 			args[argc++] = t;
 			while (*f != ' ' || in_quote) {
 				if (!*f) {
@@ -401,16 +417,23 @@ static pid_t do_cmd(char *cmd, char *machine, char *user, char *path,
 
 		server_options(args,&argc);
 
-		if (argc >= MAX_ARGS - 2) {
-			rprintf(FERROR, "internal: args[] overflowed in do_cmd()\n");
-			exit_cleanup(RERR_SYNTAX);
-		}
+		if (argc >= MAX_ARGS - 2)
+			goto arg_overflow;
 	}
 
 	args[argc++] = ".";
 
-	if (!daemon_over_rsh && path && *path)
-		args[argc++] = path;
+	if (!daemon_over_rsh) {
+		while (remote_argc > 0) {
+			if (argc >= MAX_ARGS - 1) {
+			  arg_overflow:
+				rprintf(FERROR, "internal: args[] overflowed in do_cmd()\n");
+				exit_cleanup(RERR_SYNTAX);
+			}
+			args[argc++] = *remote_argv++;
+			remote_argc--;
+		}
+	}
 
 	args[argc] = NULL;
 
@@ -422,25 +445,70 @@ static pid_t do_cmd(char *cmd, char *machine, char *user, char *path,
 
 	if (read_batch) {
 		int from_gen_pipe[2];
+		set_allow_inc_recurse();
 		if (fd_pair(from_gen_pipe) < 0) {
 			rsyserr(FERROR, errno, "pipe");
 			exit_cleanup(RERR_IPC);
 		}
 		batch_gen_fd = from_gen_pipe[0];
-		*f_out = from_gen_pipe[1];
-		*f_in = batch_fd;
+		*f_out_p = from_gen_pipe[1];
+		*f_in_p = batch_fd;
 		ret = -1; /* no child pid */
+#ifdef ICONV_CONST
+		setup_iconv();
+#endif
 	} else if (local_server) {
 		/* If the user didn't request --[no-]whole-file, force
 		 * it on, but only if we're not batch processing. */
 		if (whole_file < 0 && !write_batch)
 			whole_file = 1;
-		ret = local_child(argc, args, f_in, f_out, child_main);
-	} else
-		ret = piped_child(args,f_in,f_out);
+		set_allow_inc_recurse();
+		ret = local_child(argc, args, f_in_p, f_out_p, child_main);
+#ifdef ICONV_CONST
+		setup_iconv();
+#endif
+	} else {
+#ifdef ICONV_CONST
+		setup_iconv();
+#endif
+		if (protect_args) {
+			int fd;
+#ifdef ICONV_OPTION
+			int convert = ic_send != (iconv_t)-1;
+			xbuf outbuf, inbuf;
 
-	if (dir)
-		free(dir);
+			if (convert)
+				alloc_xbuf(&outbuf, 1024);
+#endif
+
+			ret = piped_child(args, f_in_p, f_out_p);
+
+			for (i = 0; args[i]; i++) {} /* find first NULL */
+			args[i] = "rsync"; /* set a new arg0 */
+			if (verbose > 1)
+				print_child_argv("protected args:", args + i + 1);
+			fd = *f_out_p;
+			do {
+#ifdef ICONV_OPTION
+				if (convert) {
+					INIT_XBUF_STRLEN(inbuf, args[i]);
+					iconvbufs(ic_send, &inbuf, &outbuf,
+						  ICB_EXPAND_OUT | ICB_INCLUDE_BAD | ICB_INCLUDE_INCOMPLETE);
+					outbuf.buf[outbuf.len] = '\0';
+					write_buf(fd, outbuf.buf, outbuf.len + 1);
+					outbuf.len = 0;
+				} else
+#endif
+					write_buf(fd, args[i], strlen(args[i]) + 1);
+			} while (args[++i]);
+			write_byte(fd, 0);
+#ifdef ICONV_OPTION
+			if (convert)
+				free(outbuf.buf);
+#endif
+		} else
+			ret = piped_child(args, f_in_p, f_out_p);
+	}
 
 	return ret;
 
@@ -470,7 +538,7 @@ static char *get_local_name(struct file_list *flist, char *dest_path)
 
 	if (verbose > 2) {
 		rprintf(FINFO, "get_local_name count=%d %s\n",
-			flist->count, NS(dest_path));
+			file_total, NS(dest_path));
 	}
 
 	if (!dest_path || list_only)
@@ -487,14 +555,13 @@ static char *get_local_name(struct file_list *flist, char *dest_path)
 			}
 			return NULL;
 		}
-		if (flist->count > 1) {
+		if (file_total > 1) {
 			rprintf(FERROR,
 				"ERROR: destination must be a directory when"
 				" copying more than 1 file\n");
 			exit_cleanup(RERR_FILESELECT);
 		}
-		/* Caution: flist->count could be 0! */
-		if (flist->count == 1 && S_ISDIR(flist->files[0]->mode)) {
+		if (file_total == 1 && S_ISDIR(flist->files[0]->mode)) {
 			rprintf(FERROR,
 				"ERROR: cannot overwrite non-directory"
 				" with a directory\n");
@@ -512,7 +579,7 @@ static char *get_local_name(struct file_list *flist, char *dest_path)
 	/* If we need a destination directory because the transfer is not
 	 * of a single non-directory or the user has requested one via a
 	 * destination path ending in a slash, create one and use mode 1. */
-	if (flist->count > 1 || (cp && !cp[1])) {
+	if (file_total > 1 || (cp && !cp[1])) {
 		/* Lop off the final slash (if any). */
 		if (cp && !cp[1])
 			*cp = '\0';
@@ -529,7 +596,9 @@ static char *get_local_name(struct file_list *flist, char *dest_path)
 			exit_cleanup(RERR_FILEIO);
 		}
 
-		new_root_dir = 1;
+		if (flist->high >= flist->low
+		 && strcmp(flist->files[flist->low]->basename, ".") == 0)
+			flist->files[0]->flags |= FLAG_DIR_CREATED;
 
 		if (verbose)
 			rprintf(FINFO, "created directory %s\n", dest_path);
@@ -569,63 +638,65 @@ static char *get_local_name(struct file_list *flist, char *dest_path)
 	return cp + 1;
 }
 
-/* Call this if the destination dir (which is assumed to be in curr_dir)
- * does not yet exist and we can't create it due to being in dry-run
- * mode.  We'll fix dirs that can be relative to the non-existent dir. */
-static void fix_basis_dirs(void)
+/* This function checks on our alternate-basis directories.  If we're in
+ * dry-run mode and the destination dir does not yet exist, we'll try to
+ * tweak any dest-relative paths to make them work for a dry-run (the
+ * destination dir must be in curr_dir[] when this function is called).
+ * We also warn about any arg that is non-existent or not a directory. */
+static void check_alt_basis_dirs(void)
 {
-	char **dir, *new, *slash;
-	int len;
+	STRUCT_STAT st;
+	char **dir_p, *slash = strrchr(curr_dir, '/');
 
-	if (dry_run <= 1)
-		return;
-
-	slash = strrchr(curr_dir, '/');
-
-	for (dir = basis_dir; *dir; dir++) {
-		if (**dir == '/')
-			continue;
-		len = curr_dir_len + 1 + strlen(*dir) + 1;
-		if (!(new = new_array(char, len)))
-			out_of_memory("fix_basis_dirs");
-		if (slash && strncmp(*dir, "../", 3) == 0) {
-		    /* We want to remove only one leading "../" prefix for
-		     * the directory we couldn't create in dry-run mode:
-		     * this ensures that any other ".." references get
-		     * evaluated the same as they would for a live copy. */
-		    *slash = '\0';
-		    pathjoin(new, len, curr_dir, *dir + 3);
-		    *slash = '/';
-		} else
-		    pathjoin(new, len, curr_dir, *dir);
-		*dir = new;
+	for (dir_p = basis_dir; *dir_p; dir_p++) {
+		if (dry_run > 1 && **dir_p != '/') {
+			int len = curr_dir_len + 1 + strlen(*dir_p) + 1;
+			char *new = new_array(char, len);
+			if (!new)
+				out_of_memory("check_alt_basis_dirs");
+			if (slash && strncmp(*dir_p, "../", 3) == 0) {
+			    /* We want to remove only one leading "../" prefix for
+			     * the directory we couldn't create in dry-run mode:
+			     * this ensures that any other ".." references get
+			     * evaluated the same as they would for a live copy. */
+			    *slash = '\0';
+			    pathjoin(new, len, curr_dir, *dir_p + 3);
+			    *slash = '/';
+			} else
+			    pathjoin(new, len, curr_dir, *dir_p);
+			*dir_p = new;
+		}
+		if (do_stat(*dir_p, &st) < 0) {
+			rprintf(FWARNING, "%s arg does not exist: %s\n",
+				dest_option, *dir_p);
+		} else if (!S_ISDIR(st.st_mode)) {
+			rprintf(FWARNING, "%s arg is not a dir: %s\n",
+				dest_option, *dir_p);
+		}
 	}
 }
 
 /* This is only called by the sender. */
-static void read_final_goodbye(int f_in, int f_out)
+static void read_final_goodbye(int f_in)
 {
-	int i;
+	int i, iflags, xlen;
+	uchar fnamecmp_type;
+	char xname[MAXPATHLEN];
 
 	if (protocol_version < 29)
 		i = read_int(f_in);
 	else {
-		while ((i = read_int(f_in)) == the_file_list->count
-		    && read_shortint(f_in) == ITEM_IS_NEW) {
-			/* Forward the keep-alive (no-op) to the receiver. */
-			write_int(f_out, the_file_list->count);
-			write_shortint(f_out, ITEM_IS_NEW);
-		}
+		i = read_ndx_and_attrs(f_in, &iflags, &fnamecmp_type,
+				       xname, &xlen);
 	}
 
-	if (i != -1) {
+	if (i != NDX_DONE) {
 		rprintf(FERROR, "Invalid packet at end of run (%d) [%s]\n",
 			i, who_am_i());
 		exit_cleanup(RERR_PROTOCOL);
 	}
 }
 
-
 static void do_server_sender(int f_in, int f_out, int argc, char *argv[])
 {
 	struct file_list *flist;
@@ -659,32 +730,29 @@ static void do_server_sender(int f_in, int f_out, int argc, char *argv[])
 	argc--;
 	argv++;
 
-	if (argc == 0 && (recurse || list_only)) {
+	if (argc == 0 && (recurse || xfer_dirs || list_only)) {
 		argc = 1;
 		argv--;
 		argv[0] = ".";
 	}
 
 	flist = send_file_list(f_out,argc,argv);
-	if (!flist || flist->count == 0) {
+	if (!flist || flist->used == 0)
 		exit_cleanup(0);
-	}
-	the_file_list = flist;
 
-	io_start_buffering_in();
-	io_start_buffering_out();
+	io_start_buffering_in(f_in);
 
-	send_files(flist,f_out,f_in);
+	send_files(f_in, f_out);
 	io_flush(FULL_FLUSH);
 	handle_stats(f_out);
 	if (protocol_version >= 24)
-		read_final_goodbye(f_in, f_out);
+		read_final_goodbye(f_in);
 	io_flush(FULL_FLUSH);
 	exit_cleanup(0);
 }
 
 
-static int do_recv(int f_in,int f_out,struct file_list *flist,char *local_name)
+static int do_recv(int f_in, int f_out, char *local_name)
 {
 	int pid;
 	int exit_code = 0;
@@ -692,10 +760,12 @@ static int do_recv(int f_in,int f_out,struct file_list *flist,char *local_name)
 
 	/* The receiving side mustn't obey this, or an existing symlink that
 	 * points to an identical file won't be replaced by the referent. */
-	copy_links = copy_dirlinks = 0;
+	copy_links = copy_dirlinks = copy_unsafe_links = 0;
 
-	if (preserve_hard_links)
-		init_hard_links();
+#ifdef SUPPORT_HARD_LINKS
+	if (preserve_hard_links && !inc_recurse)
+		match_hard_links(first_flist);
+#endif
 
 	if (fd_pair(error_pipe) < 0) {
 		rsyserr(FERROR, errno, "pipe failed in do_recv");
@@ -715,26 +785,32 @@ static int do_recv(int f_in,int f_out,struct file_list *flist,char *local_name)
 			close(f_out);
 
 		/* we can't let two processes write to the socket at one time */
-		close_multiplexing_out();
+		io_end_multiplex_out();
 
 		/* set place to send errors */
 		set_msg_fd_out(error_pipe[1]);
+		io_start_buffering_out(error_pipe[1]);
 
-		recv_files(f_in, flist, local_name);
+		recv_files(f_in, local_name);
 		io_flush(FULL_FLUSH);
 		handle_stats(f_in);
 
-		send_msg(MSG_DONE, "", 0);
+		send_msg(MSG_DONE, "", 1, 0);
+		write_varlong(error_pipe[1], stats.total_read, 3);
 		io_flush(FULL_FLUSH);
 
 		/* Handle any keep-alive packets from the post-processing work
 		 * that the generator does. */
 		if (protocol_version >= 29) {
+			int iflags, xlen;
+			uchar fnamecmp_type;
+			char xname[MAXPATHLEN];
+
 			kluge_around_eof = -1;
 
 			/* This should only get stopped via a USR2 signal. */
-			while (read_int(f_in) == flist->count
-			    && read_shortint(f_in) == ITEM_IS_NEW) {}
+			read_ndx_and_attrs(f_in, &iflags, &fnamecmp_type,
+					   xname, &xlen);
 
 			rprintf(FERROR, "Invalid packet at end of run [%s]\n",
 				who_am_i());
@@ -749,7 +825,8 @@ static int do_recv(int f_in,int f_out,struct file_list *flist,char *local_name)
 	}
 
 	am_generator = 1;
-	close_multiplexing_in();
+
+	io_end_multiplex_in();
 	if (write_batch && !am_server)
 		stop_write_batch();
 
@@ -757,17 +834,26 @@ static int do_recv(int f_in,int f_out,struct file_list *flist,char *local_name)
 	if (f_in != f_out)
 		close(f_in);
 
-	io_start_buffering_out();
+	io_start_buffering_out(f_out);
 
 	set_msg_fd_in(error_pipe[0]);
+	io_start_buffering_in(error_pipe[0]);
 
-	generate_files(f_out, flist, local_name);
+#ifdef SUPPORT_HARD_LINKS
+	if (preserve_hard_links && inc_recurse) {
+		struct file_list *flist;
+		for (flist = first_flist; flist; flist = flist->next)
+			match_hard_links(flist);
+	}
+#endif
+
+	generate_files(f_out, local_name);
 
 	handle_stats(-1);
 	io_flush(FULL_FLUSH);
 	if (protocol_version >= 24) {
 		/* send a final goodbye message */
-		write_int(f_out, -1);
+		write_ndx(f_out, NDX_DONE);
 	}
 	io_flush(FULL_FLUSH);
 
@@ -777,13 +863,11 @@ static int do_recv(int f_in,int f_out,struct file_list *flist,char *local_name)
 	return exit_code;
 }
 
-
-static void do_server_recv(int f_in, int f_out, int argc,char *argv[])
+static void do_server_recv(int f_in, int f_out, int argc, char *argv[])
 {
 	int exit_code;
 	struct file_list *flist;
 	char *local_name = NULL;
-	char *dir = NULL;
 	int save_verbose = verbose;
 
 	if (filesfrom_fd >= 0) {
@@ -804,7 +888,7 @@ static void do_server_recv(int f_in, int f_out, int argc,char *argv[])
 	}
 
 	if (argc > 0) {
-		dir = argv[0];
+		char *dir = argv[0];
 		argc--;
 		argv++;
 		if (!am_daemon && !push_dir(dir, 0)) {
@@ -814,7 +898,10 @@ static void do_server_recv(int f_in, int f_out, int argc,char *argv[])
 		}
 	}
 
-	io_start_buffering_in();
+	if (protocol_version >= 30)
+		io_start_multiplex_in();
+	else
+		io_start_buffering_in(f_in);
 	recv_filter_list(f_in);
 
 	if (filesfrom_fd >= 0) {
@@ -828,12 +915,13 @@ static void do_server_recv(int f_in, int f_out, int argc,char *argv[])
 	}
 
 	flist = recv_file_list(f_in);
-	verbose = save_verbose;
 	if (!flist) {
 		rprintf(FERROR,"server_recv: recv_file_list error\n");
 		exit_cleanup(RERR_FILESELECT);
 	}
-	the_file_list = flist;
+	if (inc_recurse && file_total == 1)
+		recv_additional_file_list(f_in);
+	verbose = save_verbose;
 
 	if (argc > 0)
 		local_name = get_local_name(flist,argv[0]);
@@ -841,26 +929,27 @@ static void do_server_recv(int f_in, int f_out, int argc,char *argv[])
 	/* Now that we know what our destination directory turned out to be,
 	 * we can sanitize the --link-/copy-/compare-dest args correctly. */
 	if (sanitize_paths) {
-		char **dir;
-		for (dir = basis_dir; *dir; dir++) {
-			*dir = sanitize_path(NULL, *dir, NULL, curr_dir_depth, NULL);
-		}
-		if (partial_dir) {
-			partial_dir = sanitize_path(NULL, partial_dir, NULL, curr_dir_depth, NULL);
-		}
+		char **dir_p;
+		for (dir_p = basis_dir; *dir_p; dir_p++)
+			*dir_p = sanitize_path(NULL, *dir_p, NULL, curr_dir_depth);
+		if (partial_dir)
+			partial_dir = sanitize_path(NULL, partial_dir, NULL, curr_dir_depth);
 	}
-	fix_basis_dirs();
+	check_alt_basis_dirs();
 
 	if (server_filter_list.head) {
-		char **dir;
+		char **dir_p;
 		struct filter_list_struct *elp = &server_filter_list;
 
-		for (dir = basis_dir; *dir; dir++) {
-			if (check_filter(elp, *dir, 1) < 0)
+		for (dir_p = basis_dir; *dir_p; dir_p++) {
+			char *dir = *dir_p;
+			if (*dir == '/')
+				dir += module_dirlen;
+			if (check_filter(elp, dir, 1) < 0)
 				goto options_rejected;
 		}
 		if (partial_dir && *partial_dir == '/'
-		 && check_filter(elp, partial_dir, 1) < 0) {
+		 && check_filter(elp, partial_dir + module_dirlen, 1) < 0) {
 		    options_rejected:
 			rprintf(FERROR,
 				"Your options have been rejected by the server.\n");
@@ -868,7 +957,7 @@ static void do_server_recv(int f_in, int f_out, int argc,char *argv[])
 		}
 	}
 
-	exit_code = do_recv(f_in,f_out,flist,local_name);
+	exit_code = do_recv(f_in, f_out, local_name);
 	exit_cleanup(exit_code);
 }
 
@@ -887,9 +976,6 @@ void start_server(int f_in, int f_out, int argc, char *argv[])
 
 	io_set_sock_fds(f_in, f_out);
 	setup_protocol(f_out, f_in);
-#if defined HAVE_ICONV_OPEN && defined HAVE_ICONV_H
-	setup_iconv();
-#endif
 
 	if (protocol_version >= 23)
 		io_start_multiplex_out();
@@ -898,12 +984,10 @@ void start_server(int f_in, int f_out, int argc, char *argv[])
 		keep_dirlinks = 0; /* Must be disabled on the sender. */
 		if (need_messages_from_generator)
 			io_start_multiplex_in();
-
 		recv_filter_list(f_in);
 		do_server_sender(f_in, f_out, argc, argv);
-	} else {
+	} else
 		do_server_recv(f_in, f_out, argc, argv);
-	}
 	exit_cleanup(0);
 }
 
@@ -926,12 +1010,6 @@ int client_run(int f_in, int f_out, pid_t pid, int argc, char *argv[])
 
 	io_set_sock_fds(f_in, f_out);
 	setup_protocol(f_out,f_in);
-#if defined HAVE_ICONV_OPEN && defined HAVE_ICONV_H
-	setup_iconv();
-#endif
-
-	if (protocol_version >= 23 && !read_batch)
-		io_start_multiplex_in();
 
 	/* We set our stderr file handle to blocking because ssh might have
 	 * set it to non-blocking.  This can be particularly troublesome if
@@ -945,7 +1023,10 @@ int client_run(int f_in, int f_out, pid_t pid, int argc, char *argv[])
 
 	if (am_sender) {
 		keep_dirlinks = 0; /* Must be disabled on the sender. */
-		io_start_buffering_out();
+		if (protocol_version >= 30)
+			io_start_multiplex_out();
+		else
+			io_start_buffering_out(f_out);
 		if (!filesfrom_host)
 			set_msg_fd_in(f_in);
 		send_filter_list(f_out);
@@ -958,14 +1039,16 @@ int client_run(int f_in, int f_out, pid_t pid, int argc, char *argv[])
 		set_msg_fd_in(-1);
 		if (verbose > 3)
 			rprintf(FINFO,"file list sent\n");
-		the_file_list = flist;
+
+		if (protocol_version >= 23)
+			io_start_multiplex_in();
 
 		io_flush(NORMAL_FLUSH);
-		send_files(flist,f_out,f_in);
+		send_files(f_in, f_out);
 		io_flush(FULL_FLUSH);
 		handle_stats(-1);
 		if (protocol_version >= 24)
-			read_final_goodbye(f_in, f_out);
+			read_final_goodbye(f_in);
 		if (pid != -1) {
 			if (verbose > 3)
 				rprintf(FINFO,"client_run waiting on %d\n", (int) pid);
@@ -977,11 +1060,17 @@ int client_run(int f_in, int f_out, pid_t pid, int argc, char *argv[])
 		exit_cleanup(exit_code);
 	}
 
-	if (need_messages_from_generator && !read_batch)
-		io_start_multiplex_out();
+	if (!read_batch) {
+		if (protocol_version >= 23)
+			io_start_multiplex_in();
+		if (need_messages_from_generator)
+			io_start_multiplex_out();
+	}
 
-	if (argc == 0)
+	if (argc == 0) {
 		list_only |= 1;
+		xfer_dirs |= 1;
+	}
 
 	send_filter_list(read_batch ? -1 : f_out);
 
@@ -993,14 +1082,15 @@ int client_run(int f_in, int f_out, pid_t pid, int argc, char *argv[])
 	if (write_batch && !am_server)
 		start_write_batch(f_in);
 	flist = recv_file_list(f_in);
-	the_file_list = flist;
+	if (inc_recurse && file_total == 1)
+		recv_additional_file_list(f_in);
 
-	if (flist && flist->count > 0) {
+	if (flist && flist->used > 0) {
 		local_name = get_local_name(flist, argv[0]);
 
-		fix_basis_dirs();
+		check_alt_basis_dirs();
 
-		exit_code2 = do_recv(f_in, f_out, flist, local_name);
+		exit_code2 = do_recv(f_in, f_out, local_name);
 	} else {
 		handle_stats(-1);
 		output_summary();
@@ -1016,7 +1106,7 @@ int client_run(int f_in, int f_out, pid_t pid, int argc, char *argv[])
 	return MAX(exit_code, exit_code2);
 }
 
-static int copy_argv (char *argv[])
+static int copy_argv(char *argv[])
 {
 	int i;
 
@@ -1042,66 +1132,69 @@ static int copy_argv (char *argv[])
  **/
 static int start_client(int argc, char *argv[])
 {
-	char *p;
-	char *shell_machine = NULL;
-	char *shell_path = NULL;
-	char *shell_user = NULL;
+	char *p, *shell_machine = NULL, *shell_user = NULL;
+	char **remote_argv;
+	int remote_argc;
+	int f_in, f_out;
 	int ret;
 	pid_t pid;
-	int f_in,f_out;
-	int rc;
 
 	/* Don't clobber argv[] so that ps(1) can still show the right
 	 * command line. */
-	if ((rc = copy_argv(argv)))
-		return rc;
+	if ((ret = copy_argv(argv)) != 0)
+		return ret;
 
 	if (!read_batch) { /* for read_batch, NO source is specified */
-		shell_path = check_for_hostspec(argv[0], &shell_machine, &rsync_port);
-		if (shell_path) { /* source is remote */
-			char *dummy1;
-			int dummy2;
-			if (--argc
-			 && check_for_hostspec(argv[argc], &dummy1, &dummy2)) {
+		char *path = check_for_hostspec(argv[0], &shell_machine, &rsync_port);
+		if (path) { /* source is remote */
+			char *dummy_host;
+			int dummy_port = 0;
+			*argv = path;
+			remote_argv = argv;
+			remote_argc = argc;
+			argv += argc - 1;
+			if (argc == 1 || **argv == ':')
+				argc = 0; /* no dest arg */
+			else if (check_for_hostspec(*argv, &dummy_host, &dummy_port)) {
 				rprintf(FERROR,
 					"The source and destination cannot both be remote.\n");
 				exit_cleanup(RERR_SYNTAX);
+			} else {
+				remote_argc--; /* don't count dest */
+				argc = 1;
 			}
-			argv++;
 			if (filesfrom_host && *filesfrom_host
 			    && strcmp(filesfrom_host, shell_machine) != 0) {
 				rprintf(FERROR,
 					"--files-from hostname is not the same as the transfer hostname\n");
 				exit_cleanup(RERR_SYNTAX);
 			}
-			if (rsync_port) {
-				if (!shell_cmd) {
-					return start_socket_client(shell_machine,
-								   shell_path,
-								   argc, argv);
-				}
-				daemon_over_rsh = 1;
-			}
-
 			am_sender = 0;
+			if (rsync_port)
+				daemon_over_rsh = shell_cmd ? 1 : -1;
 		} else { /* source is local, check dest arg */
 			am_sender = 1;
 
-			if (argc > 1)
+			if (argc > 1) {
 				p = argv[--argc];
-			else {
-				p = ".";
-				list_only = 1;
+				remote_argv = argv + argc;
+			} else {
+				static char *dotarg[1] = { "." };
+				p = dotarg[0];
+				remote_argv = dotarg;
+				list_only |= 1;
+				xfer_dirs |= 1;
 			}
+			remote_argc = 1;
 
-			shell_path = check_for_hostspec(p, &shell_machine, &rsync_port);
-			if (shell_path && filesfrom_host && *filesfrom_host
+			path = check_for_hostspec(p, &shell_machine, &rsync_port);
+			if (path && filesfrom_host && *filesfrom_host
 			    && strcmp(filesfrom_host, shell_machine) != 0) {
 				rprintf(FERROR,
 					"--files-from hostname is not the same as the transfer hostname\n");
 				exit_cleanup(RERR_SYNTAX);
 			}
-			if (!shell_path) { /* no hostspec found, so src & dest are local */
+			if (!path) { /* no hostspec found, so src & dest are local */
 				local_server = 1;
 				if (filesfrom_host) {
 					rprintf(FERROR,
@@ -1109,23 +1202,78 @@ static int start_client(int argc, char *argv[])
 					exit_cleanup(RERR_SYNTAX);
 				}
 				shell_machine = NULL;
-				shell_path = p;
-			} else if (rsync_port) {
-				if (!shell_cmd) {
-					return start_socket_client(shell_machine,
-								   shell_path,
-								   argc, argv);
-				}
-				daemon_over_rsh = 1;
+			} else { /* hostspec was found, so dest is remote */
+				argv[argc] = path;
+				if (rsync_port)
+					daemon_over_rsh = shell_cmd ? 1 : -1;
 			}
 		}
 	} else {  /* read_batch */
 		local_server = 1;
-		shell_path = argv[argc-1];
-		if (check_for_hostspec(shell_path, &shell_machine, &rsync_port)) {
+		if (check_for_hostspec(argv[argc-1], &shell_machine, &rsync_port)) {
 			rprintf(FERROR, "remote destination is not allowed with --read-batch\n");
 			exit_cleanup(RERR_SYNTAX);
 		}
+		remote_argv = argv + argc - 1;
+		remote_argc = 1;
+	}
+
+	if (am_sender) {
+		char *dummy_host;
+		int dummy_port = rsync_port;
+		int i;
+		/* For local source, extra source args must not have hostspec. */
+		for (i = 1; i < argc; i++) {
+			if (check_for_hostspec(argv[i], &dummy_host, &dummy_port)) {
+				rprintf(FERROR, "Unexpected remote arg: %s\n", argv[i]);
+				exit_cleanup(RERR_SYNTAX);
+			}
+		}
+	} else {
+		char *dummy_host;
+		int dummy_port = rsync_port;
+		int i;
+		/* For remote source, any extra source args must have either
+		 * the same hostname or an empty hostname. */
+		for (i = 1; i < remote_argc; i++) {
+			char *arg = check_for_hostspec(remote_argv[i], &dummy_host, &dummy_port);
+			if (!arg) {
+				rprintf(FERROR, "Unexpected local arg: %s\n", remote_argv[i]);
+				rprintf(FERROR, "If arg is a remote file/dir, prefix it with a colon (:).\n");
+				exit_cleanup(RERR_SYNTAX);
+			}
+			if (*dummy_host && strcmp(dummy_host, shell_machine) != 0) {
+				rprintf(FERROR, "All source args must come from the same machine.\n");
+				exit_cleanup(RERR_SYNTAX);
+			}
+			if (rsync_port != dummy_port) {
+				if (!rsync_port || !dummy_port)
+					rprintf(FERROR, "All source args must use the same hostspec format.\n");
+				else
+					rprintf(FERROR, "All source args must use the same port number.\n");
+				exit_cleanup(RERR_SYNTAX);
+			}
+			remote_argv[i] = arg;
+		}
+		if (argc == 0) {
+			list_only |= 1;
+			xfer_dirs |= 1;
+		}
+	}
+
+	if (daemon_over_rsh < 0)
+		return start_socket_client(shell_machine, remote_argc, remote_argv, argc, argv);
+
+	if (password_file && !daemon_over_rsh) {
+		rprintf(FERROR, "The --password-file option may only be "
+				"used when accessing an rsync daemon.\n");
+		exit_cleanup(RERR_SYNTAX);
+	}
+
+	if (connect_timeout) {
+		rprintf(FERROR, "The --contimeout option may only be "
+				"used when connecting to an rsync daemon.\n");
+		exit_cleanup(RERR_SYNTAX);
 	}
 
 	if (shell_machine) {
@@ -1139,31 +1287,18 @@ static int start_client(int argc, char *argv[])
 
 	if (verbose > 3) {
 		rprintf(FINFO,"cmd=%s machine=%s user=%s path=%s\n",
-			shell_cmd ? shell_cmd : "",
-			shell_machine ? shell_machine : "",
-			shell_user ? shell_user : "",
-			shell_path ? shell_path : "");
+			NS(shell_cmd), NS(shell_machine), NS(shell_user),
+			remote_argv ? NS(remote_argv[0]) : "");
 	}
 
-	/* for remote source, only single dest arg can remain ... */
-	if (!am_sender && argc > 1) {
-		usage(FERROR);
-		exit_cleanup(RERR_SYNTAX);
-	}
-
-	/* ... or no dest at all */
-	if (!am_sender && argc == 0)
-		list_only |= 1;
-
-	pid = do_cmd(shell_cmd,shell_machine,shell_user,shell_path,
-		     &f_in,&f_out);
+	pid = do_cmd(shell_cmd, shell_machine, shell_user, remote_argv, remote_argc,
+		     &f_in, &f_out);
 
 	/* if we're running an rsync server on the remote host over a
 	 * remote shell command, we need to do the RSYNCD protocol first */
 	if (daemon_over_rsh) {
 		int tmpret;
-		tmpret = start_inband_exchange(shell_user, shell_path,
-					       f_in, f_out, argc);
+		tmpret = start_inband_exchange(f_in, f_out, shell_user, remote_argc, remote_argv);
 		if (tmpret < 0)
 			return tmpret;
 	}
@@ -1187,7 +1322,7 @@ static RETSIGTYPE sigusr2_handler(UNUSED(int val))
 	if (!am_server)
 		output_summary();
 	close_all();
-	if (log_got_error)
+	if (got_xfer_error)
 		_exit(RERR_PARTIAL);
 	_exit(0);
 }
@@ -1294,7 +1429,8 @@ int main(int argc,char *argv[])
 #endif
 
 	starttime = time(NULL);
-	am_root = (MY_UID() == 0);
+	our_uid = MY_UID();
+	am_root = our_uid == 0;
 
 	memset(&stats, 0, sizeof(stats));
 
@@ -1358,13 +1494,30 @@ int main(int argc,char *argv[])
 		}
 		if (read_batch)
 			read_stream_flags(batch_fd);
+		else
+			write_stream_flags(batch_fd);
+
 	}
 	if (write_batch < 0)
 		dry_run = 1;
 
-	if (am_daemon && !am_server)
+	if (am_server) {
+#ifdef ICONV_CONST
+		setup_iconv();
+#endif
+	} else if (am_daemon)
 		return daemon_main();
 
+	if (am_server && protect_args) {
+		char buf[MAXPATHLEN];
+		protect_args = 2;
+		read_args(STDIN_FILENO, NULL, buf, sizeof buf, 1, &argv, &argc, NULL);
+		if (!parse_arguments(&argc, (const char ***) &argv, 1)) {
+			option_error();
+			exit_cleanup(RERR_SYNTAX);
+		}
+	}
+
 	if (argc < 1) {
 		usage(FERROR);
 		exit_cleanup(RERR_SYNTAX);
