@@ -2,11 +2,11 @@
  * Copyright (C) 1996, 2000 Andrew Tridgell
  * Copyright (C) 1996 Paul Mackerras
  * Copyright (C) 2001, 2002 Martin Pool <mbp@samba.org>
- * Copyright (C) 2003, 2004, 2005, 2006 Wayne Davison
+ * Copyright (C) 2003-2008 Wayne Davison
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -15,8 +15,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
+ * with this program; if not, visit the http://fsf.org website.
  */
 
 #define False 0
@@ -33,6 +32,9 @@
 #define DEFAULT_LOCK_FILE "/var/run/rsyncd.lock"
 #define URL_PREFIX "rsync://"
 
+#define SYMLINK_PREFIX "/rsyncd-munged/"
+#define SYMLINK_PREFIX_LEN ((int)sizeof SYMLINK_PREFIX - 1)
+
 #define BACKUP_SUFFIX "~"
 
 /* a non-zero CHAR_OFFSET makes the rolling sum stronger, but is
@@ -43,31 +45,56 @@
 
 #define XMIT_TOP_DIR (1<<0)
 #define XMIT_SAME_MODE (1<<1)
-#define XMIT_EXTENDED_FLAGS (1<<2)
-#define XMIT_SAME_RDEV_pre28 XMIT_EXTENDED_FLAGS /* Only in protocols < 28 */
+#define XMIT_SAME_RDEV_pre28 (1<<2)	/* protocols 20 - 27  */
+#define XMIT_EXTENDED_FLAGS (1<<2)	/* protocols 28 - now */
 #define XMIT_SAME_UID (1<<3)
 #define XMIT_SAME_GID (1<<4)
 #define XMIT_SAME_NAME (1<<5)
 #define XMIT_LONG_NAME (1<<6)
 #define XMIT_SAME_TIME (1<<7)
-#define XMIT_SAME_RDEV_MAJOR (1<<8)
-#define XMIT_HAS_IDEV_DATA (1<<9)
-#define XMIT_SAME_DEV (1<<10)
-#define XMIT_RDEV_MINOR_IS_SMALL (1<<11)
+#define XMIT_SAME_RDEV_MAJOR (1<<8)	/* protocols 28 - now (devices only) */
+#define XMIT_NO_CONTENT_DIR (1<<8)	/* protocols 30 - now (dirs only) */
+#define XMIT_HLINKED (1<<9)		/* protocols 28 - now */
+#define XMIT_SAME_DEV_pre30 (1<<10)	/* protocols 28 - 29  */
+#define XMIT_USER_NAME_FOLLOWS (1<<10)	/* protocols 30 - now */
+#define XMIT_RDEV_MINOR_8_pre30 (1<<11)	/* protocols 28 - 29  */
+#define XMIT_GROUP_NAME_FOLLOWS (1<<11) /* protocols 30 - now */
+#define XMIT_HLINK_FIRST (1<<12)	/* protocols 30 - now (HLINKED files only) */
 
 /* These flags are used in the live flist data. */
 
-#define FLAG_TOP_DIR (1<<0)
-#define FLAG_SENT (1<<1)	/* sender */
-#define FLAG_HLINK_EOL (1<<1)	/* receiver/generator */
-#define FLAG_MOUNT_POINT (1<<2)	/* sender/generator */
-#define FLAG_DEL_HERE (1<<3)	/* receiver/generator */
-#define FLAG_HLINK_TOL (1<<4)	/* receiver/generator */
-#define FLAG_NO_FUZZY (1<<5)	/* generator */
-#define FLAG_MISSING (1<<6)	/* generator */
+#define FLAG_TOP_DIR (1<<0)	/* sender/receiver/generator */
+#define FLAG_FILE_SENT (1<<1)	/* sender/receiver/generator */
+#define FLAG_DIR_CREATED (1<<1)	/* generator */
+#define FLAG_CONTENT_DIR (1<<2)	/* sender/receiver/generator */
+#define FLAG_MOUNT_DIR (1<<3)	/* sender/generator */
+#define FLAG_DUPLICATE (1<<4)	/* sender */
+#define FLAG_MISSING_DIR (1<<4)	/* generator */
+#define FLAG_HLINKED (1<<5)	/* receiver/generator (checked on all types) */
+#define FLAG_HLINK_FIRST (1<<6)	/* receiver/generator (w/FLAG_HLINKED) */
+#define FLAG_IMPLIED_DIR (1<<6)	/* sender/receiver/generator (dirs only) */
+#define FLAG_HLINK_LAST (1<<7)	/* receiver/generator */
+#define FLAG_HLINK_DONE (1<<8)	/* receiver/generator (checked on all types) */
+#define FLAG_LENGTH64 (1<<9)	/* sender/receiver/generator */
+#define FLAG_SKIP_GROUP (1<<10)	/* receiver/generator */
+#define FLAG_TIME_FAILED (1<<11)/* generator */
+
+/* These flags are passed to functions but not stored. */
+
+#define FLAG_DIVERT_DIRS (1<<16)/* sender */
+
+#define BITS_SET(val,bits) (((val) & (bits)) == (bits))
+#define BITS_SETnUNSET(val,onbits,offbits) (((val) & ((onbits)|(offbits))) == (onbits))
+#define BITS_EQUAL(b1,b2,mask) (((unsigned)(b1) & (unsigned)(mask)) \
+			     == ((unsigned)(b2) & (unsigned)(mask)))
 
 /* update this if you make incompatible changes */
-#define PROTOCOL_VERSION 29
+#define PROTOCOL_VERSION 30
+
+/* This is used when working on a new protocol version in CVS, and should
+ * be a new non-zero value for each CVS change that affects the protocol.
+ * It must ALWAYS be 0 when the protocol goes final! */
+#define SUBPROTOCOL_VERSION 0
 
 /* We refuse to interoperate with versions that are not in this range.
  * Note that we assume we'll work with later versions: the onus is on
@@ -90,6 +117,8 @@
 #define OLD_PROTOCOL_VERSION 25
 #define MAX_PROTOCOL_VERSION 40
 
+#define FILECNT_LOOKAHEAD 1000
+
 #define RSYNC_PORT 873
 
 #define SPARSE_WRITE_SIZE (1024)
@@ -97,7 +126,7 @@
 #define CHUNK_SIZE (32*1024)
 #define MAX_MAP_SIZE (256*1024)
 #define IO_BUFFER_SIZE (4092)
-#define MAX_BLOCK_SIZE ((int32)1 << 29)
+#define MAX_BLOCK_SIZE ((int32)1 << 17)
 
 #define IOERR_GENERAL	(1<<0) /* For backward compatibility, this must == 1 */
 #define IOERR_VANISHED	(1<<1)
@@ -153,28 +182,49 @@
 /* These are outside the range of the transmitted flags. */
 #define ITEM_MISSING_DATA (1<<16)	   /* used by log_formatted() */
 #define ITEM_DELETED (1<<17)		   /* used by log_formatted() */
+#define ITEM_MATCHED (1<<18)		   /* used by itemize() */
 
 #define SIGNIFICANT_ITEM_FLAGS (~(\
 	ITEM_BASIS_TYPE_FOLLOWS | ITEM_XNAME_FOLLOWS | ITEM_LOCAL_CHANGE))
 
+#define CFN_KEEP_LEADING_DOT_DIR (1<<0)
+#define CFN_KEEP_TRAILING_SLASH (1<<1)
+#define CFN_DROP_TRAILING_DOT_DIR (1<<2)
+#define CFN_COLLAPSE_DOT_DOT_DIRS (1<<3)
 
-/* Log-message categories.  Only FERROR and FINFO get sent over the socket,
- * but FLOG and FSOCKERR can be sent over the receiver -> generator pipe.
- * FLOG only goes to the log file, not the client; FCLIENT is the opposite. */
-enum logcode { FNONE=0, FERROR=1, FINFO=2, FLOG=3, FCLIENT=4, FSOCKERR=5 };
+/* Log-message categories.  FLOG only goes to the log file, not the client;
+ * FCLIENT is the opposite. */
+enum logcode {
+    FNONE=0, /* never sent */
+    FERROR_XFER=1, FINFO=2, /* sent over socket for any protocol */
+    FERROR=3, FWARNING=4, /* sent over socket for protocols >= 30 */
+    FERROR_SOCKET=5, FLOG=6, /* only sent via receiver -> generator pipe */
+    FCLIENT=7 /* never transmitted (e.g. server converts to FINFO) */
+};
 
 /* Messages types that are sent over the message channel.  The logcode
  * values must all be present here with identical numbers. */
 enum msgcode {
 	MSG_DATA=0,	/* raw data on the multiplexed stream */
-	MSG_ERROR=FERROR, MSG_INFO=FINFO, /* remote logging */
-	MSG_LOG=FLOG, MSG_SOCKERR=FSOCKERR, /* sibling logging */
+	MSG_ERROR_XFER=FERROR_XFER, MSG_INFO=FINFO, /* remote logging */
+	MSG_ERROR=FERROR, MSG_WARNING=FWARNING, /* protocol-30 remote logging */
+	MSG_ERROR_SOCKET=FERROR_SOCKET, /* sibling logging */
+	MSG_LOG=FLOG, MSG_CLIENT=FCLIENT, /* sibling logging */
 	MSG_REDO=9,	/* reprocess indicated flist index */
+	MSG_FLIST=20,	/* extra file list over sibling socket */
+	MSG_FLIST_EOF=21,/* we've transmitted all the file lists */
+	MSG_IO_ERROR=22,/* the sending side had an I/O error */
+	MSG_NOOP=42,	/* a do-nothing message */
 	MSG_SUCCESS=100,/* successfully updated indicated flist index */
 	MSG_DELETED=101,/* successfully deleted a file on receiving side */
+	MSG_NO_SEND=102,/* sender failed to open a file we wanted */
 	MSG_DONE=86	/* current phase is done */
 };
 
+#define NDX_DONE -1
+#define NDX_FLIST_EOF -2
+#define NDX_FLIST_OFFSET -101
+
 #include "errcode.h"
 
 #include "config.h"
@@ -205,6 +255,12 @@ enum msgcode {
 #ifdef HAVE_STRINGS_H
 # include <strings.h>
 #endif
+#ifdef HAVE_INTTYPES_H
+# include <inttypes.h>
+#endif
+#ifdef HAVE_STDINT_H
+# include <stdint.h>
+#endif
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
 #endif
@@ -286,7 +342,9 @@ enum msgcode {
 #include <stdarg.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#ifdef HAVE_NETDB_H
 #include <netdb.h>
+#endif
 #include <syslog.h>
 #include <sys/file.h>
 
@@ -328,10 +386,44 @@ enum msgcode {
 # include <limits.h>
 #endif
 
+#if defined HAVE_ICONV_OPEN && defined HAVE_ICONV_H
+#include <iconv.h>
+#ifndef ICONV_CONST
+#define ICONV_CONST
+#endif
+#else
+#ifdef ICONV_CONST
+#undef ICONV_CONST
+#endif
+#ifdef ICONV_OPTION
+#undef ICONV_OPTION
+#endif
+#ifdef iconv_t
+#undef iconv_t
+#endif
+#define iconv_t int
+#endif
+
 #include <assert.h>
 
 #include "lib/pool_alloc.h"
 
+#ifndef HAVE_ID_T
+typedef unsigned int id_t;
+#endif
+#ifndef HAVE_PID_T
+typedef int pid_t;
+#endif
+#ifndef HAVE_MODE_T
+typedef unsigned int mode_t;
+#endif
+#ifndef HAVE_OFF_T
+typedef long off_t;
+#endif
+#ifndef HAVE_SIZE_T
+typedef unsigned int size_t;
+#endif
+
 #define BOOL int
 
 #ifndef uchar
@@ -344,11 +436,30 @@ enum msgcode {
 #define schar char
 #endif
 
+#ifndef int16
+#if SIZEOF_INT16_T == 2
+# define int16 int16_t
+#else
+# define int16 short
+#endif
+#endif
+
+#ifndef uint16
+#if SIZEOF_UINT16_T == 2
+# define uint16 uint16_t
+#else
+# define uint16 unsigned int16
+#endif
+#endif
+
 /* Find a variable that is either exactly 32-bits or longer.
  * If some code depends on 32-bit truncation, it will need to
  * take special action in a "#if SIZEOF_INT32 > 4" section. */
 #ifndef int32
-#if SIZEOF_INT == 4
+#if SIZEOF_INT32_T == 4
+# define int32 int32_t
+# define SIZEOF_INT32 4
+#elif SIZEOF_INT == 4
 # define int32 int
 # define SIZEOF_INT32 4
 #elif SIZEOF_LONG == 4
@@ -371,7 +482,11 @@ enum msgcode {
 #endif
 
 #ifndef uint32
-#define uint32 unsigned int32
+#if SIZEOF_UINT32_T == 4
+# define uint32 uint32_t
+#else
+# define uint32 unsigned int32
+#endif
 #endif
 
 #if SIZEOF_OFF_T == 8 || !SIZEOF_OFF64_T || !defined HAVE_STRUCT_STAT64
@@ -389,7 +504,10 @@ enum msgcode {
  * to ensure that any code that really requires a 64-bit integer has
  * it (e.g. the checksum code uses two 32-bit integers for its 64-bit
  * counter). */
-#if SIZEOF_LONG == 8
+#if SIZEOF_INT64_T == 8
+# define int64 int64_t
+# define SIZEOF_INT64 8
+#elif SIZEOF_LONG == 8
 # define int64 long
 # define SIZEOF_INT64 8
 #elif SIZEOF_INT == 8
@@ -419,37 +537,27 @@ enum msgcode {
 # define SIZEOF_INT64 SIZEOF_OFF_T
 #endif
 
-/* Starting from protocol version 26, we always use 64-bit
- * ino_t and dev_t internally, even if this platform does not
- * allow files to have 64-bit inums.  That's because the
- * receiver needs to find duplicate (dev,ino) tuples to detect
- * hardlinks, and it might have files coming from a platform
- * that has 64-bit inums.
- *
- * The only exception is if we're on a platform with no 64-bit type at
- * all.
- *
- * Because we use read_longint() to get these off the wire, if you
- * transfer devices or hardlinks with dev or inum > 2**32 to a machine
- * with no 64-bit types then you will get an overflow error.  Probably
- * not many people have that combination of machines, and you can
- * avoid it by not preserving hardlinks or not transferring device
- * nodes.  It's not clear that any other behaviour is better.
- *
- * Note that if you transfer devices from a 64-bit-devt machine (say,
- * Solaris) to a 32-bit-devt machine (say, Linux-2.2/x86) then the
- * device numbers will be truncated.  But it's a kind of silly thing
- * to do anyhow.
- *
- * FIXME: I don't think the code in flist.c has ever worked on a system
- * where dev_t is a struct.
- */
-
-struct idev {
-	int64 inode;
-	int64 dev;
+struct hashtable {
+	void *nodes;
+	int32 size, entries;
+	uint32 node_size;
+	int key64;
 };
 
+struct ht_int32_node {
+	void *data;
+	int32 key;
+};
+
+struct ht_int64_node {
+	void *data;
+	int64 key;
+};
+
+#define HT_NODE(tbl, bkts, i) ((void*)((char*)(bkts) + (i)*(tbl)->node_size))
+#define HT_KEY(node, k64) ((k64)? ((struct ht_int64_node*)(node))->key \
+			 : (int64)((struct ht_int32_node*)(node))->key)
+
 #ifndef MIN
 #define MIN(a,b) ((a)<(b)?(a):(b))
 #endif
@@ -462,8 +570,6 @@ struct idev {
 #define MAXHOSTNAMELEN 256
 #endif
 
-/* the length of the md4 checksum */
-#define MD4_SUM_LENGTH 16
 #define SUM_LENGTH 16
 #define SHORT_SUM_LENGTH 2
 #define BLOCKSUM_BIAS 10
@@ -492,47 +598,104 @@ struct idev {
 #define IN_LOOPBACKNET 127
 #endif
 
-#define GID_NONE ((gid_t)-1)
+#if HAVE_UNIXWARE_ACLS|HAVE_SOLARIS_ACLS|HAVE_HPUX_ACLS
+#define ACLS_NEED_MASK 1
+#endif
 
-#define HL_CHECK_MASTER	0
-#define HL_SKIP		1
-
-struct hlink {
-	int32 next;
-	int32 hlindex;
-	unsigned short link_dest_used;
+union file_extras {
+	int32 num;
+	uint32 unum;
 };
 
-#define F_DEV	link_u.idev->dev
-#define F_INODE	link_u.idev->inode
-
-#define F_HLINDEX link_u.links->hlindex
-#define F_NEXT	link_u.links->next
-
 struct file_struct {
-	union {
-		dev_t rdev;	/* The device number, if this is a device */
-		char *sum;	/* Only a normal file can have a checksum */
-		char *link;	/* Points to symlink string, if a symlink */
-	} u;
-	OFF_T length;
-	char *basename;		/* The current item's name (AKA filename) */
-	char *dirname;		/* The directory info inside the transfer */
-	union {
-		char *root;	/* Sender-side dir info outside transfer */
-		int depth;	/* Receiver-side directory depth info */
-	} dir;
-	union {
-		struct idev *idev;
-		struct hlink *links;
-	} link_u;
-	time_t modtime;
-	uid_t uid;
-	gid_t gid;
-	mode_t mode;
-	uchar flags;	/* this item MUST remain last */
+	const char *dirname;	/* The dir info inside the transfer */
+	time_t modtime;		/* When the item was last modified */
+	uint32 len32;		/* Lowest 32 bits of the file's length */
+	uint16 mode;		/* The item's type and permissions */
+	uint16 flags;		/* The FLAG_* bits for this item */
+	const char basename[1];	/* The basename (AKA filename) follows */
 };
 
+extern int file_extra_cnt;
+extern int inc_recurse;
+extern int uid_ndx;
+extern int gid_ndx;
+extern int acls_ndx;
+extern int xattrs_ndx;
+
+#define FILE_STRUCT_LEN (offsetof(struct file_struct, basename))
+#define EXTRA_LEN (sizeof (union file_extras))
+#define PTR_EXTRA_CNT ((sizeof (char *) + EXTRA_LEN - 1) / EXTRA_LEN)
+#define DEV_EXTRA_CNT 2
+#define DIRNODE_EXTRA_CNT 3
+#define SUM_EXTRA_CNT ((MAX_DIGEST_LEN + EXTRA_LEN - 1) / EXTRA_LEN)
+
+#define REQ_EXTRA(f,ndx) ((union file_extras*)(f) - (ndx))
+#define OPT_EXTRA(f,bump) ((union file_extras*)(f) - file_extra_cnt - 1 - (bump))
+
+#define LEN64_BUMP(f) ((f)->flags & FLAG_LENGTH64 ? 1 : 0)
+#define HLINK_BUMP(f) ((f)->flags & (FLAG_HLINKED|FLAG_HLINK_DONE) ? inc_recurse+1 : 0)
+#define ACL_BUMP(f) (acls_ndx ? 1 : 0)
+
+/* The length applies to all items. */
+#if SIZEOF_INT64 < 8
+#define F_LENGTH(f) ((int64)(f)->len32)
+#else
+#define F_LENGTH(f) ((int64)(f)->len32 + ((f)->flags & FLAG_LENGTH64 \
+		   ? (int64)OPT_EXTRA(f, 0)->unum << 32 : 0))
+#endif
+
+/* If there is a symlink string, it is always right after the basename */
+#define F_SYMLINK(f) ((f)->basename + strlen((f)->basename) + 1)
+
+/* The sending side always has this available: */
+#define F_PATHNAME(f) (*(const char**)REQ_EXTRA(f, PTR_EXTRA_CNT))
+
+/* The receiving side always has this available: */
+#define F_DEPTH(f) REQ_EXTRA(f, 1)->num
+
+/* When the associated option is on, all entries will have these present: */
+#define F_OWNER(f) REQ_EXTRA(f, uid_ndx)->unum
+#define F_GROUP(f) REQ_EXTRA(f, gid_ndx)->unum
+#define F_ACL(f) REQ_EXTRA(f, acls_ndx)->num
+#define F_XATTR(f) REQ_EXTRA(f, xattrs_ndx)->num
+#define F_NDX(f) REQ_EXTRA(f, unsort_ndx)->num
+
+/* These items are per-entry optional: */
+#define F_HL_GNUM(f) OPT_EXTRA(f, LEN64_BUMP(f))->num /* non-dirs */
+#define F_HL_PREV(f) OPT_EXTRA(f, LEN64_BUMP(f)+inc_recurse)->num /* non-dirs */
+#define F_DIR_NODE_P(f) (&OPT_EXTRA(f, LEN64_BUMP(f) \
+				+ DIRNODE_EXTRA_CNT - 1)->num) /* sender dirs */
+#define F_DIR_RELNAMES_P(f) (&OPT_EXTRA(f, LEN64_BUMP(f) + DIRNODE_EXTRA_CNT \
+				+ PTR_EXTRA_CNT - 1)->num) /* sender dirs */
+#define F_DIR_DEFACL(f) OPT_EXTRA(f, LEN64_BUMP(f))->unum /* receiver dirs */
+#define F_DIR_DEV_P(f) (&OPT_EXTRA(f, LEN64_BUMP(f) + ACL_BUMP(f) \
+				+ DEV_EXTRA_CNT - 1)->unum) /* receiver dirs */
+
+/* This optional item might follow an F_HL_*() item.
+ * (Note: a device doesn't need to check LEN64_BUMP(f).) */
+#define F_RDEV_P(f) (&OPT_EXTRA(f, HLINK_BUMP(f) + DEV_EXTRA_CNT - 1)->unum)
+
+/* The sum is only present on regular files. */
+#define F_SUM(f) ((char*)OPT_EXTRA(f, LEN64_BUMP(f) + HLINK_BUMP(f) \
+				    + SUM_EXTRA_CNT - 1))
+
+/* Some utility defines: */
+#define F_IS_ACTIVE(f) (f)->basename[0]
+#define F_IS_HLINKED(f) ((f)->flags & FLAG_HLINKED)
+
+#define F_HLINK_NOT_FIRST(f) BITS_SETnUNSET((f)->flags, FLAG_HLINKED, FLAG_HLINK_FIRST)
+#define F_HLINK_NOT_LAST(f) BITS_SETnUNSET((f)->flags, FLAG_HLINKED, FLAG_HLINK_LAST)
+
+/* These access the F_DIR_DEV_P() and F_RDEV_P() values: */
+#define DEV_MAJOR(a) (a)[0]
+#define DEV_MINOR(a) (a)[1]
+
+/* These access the F_DIRS_NODE_P() values: */
+#define DIR_PARENT(a) (a)[0]
+#define DIR_FIRST_CHILD(a) (a)[1]
+#define DIR_NEXT_SIBLING(a) (a)[2]
+
 /*
  * Start the flist array at FLIST_START entries and grow it
  * by doubling until FLIST_LINEAR then grow by FLIST_LINEAR
@@ -541,27 +704,30 @@ struct file_struct {
 #define FLIST_LINEAR	(FLIST_START * 512)
 
 /*
- * Extent size for allocation pools A minimum size of 128KB
+ * Extent size for allocation pools: A minimum size of 128KB
  * is needed to mmap them so that freeing will release the
  * space to the OS.
  *
  * Larger sizes reduce leftover fragments and speed free calls
- * (when they happen) Smaller sizes increase the chance of
+ * (when they happen). Smaller sizes increase the chance of
  * freed allocations freeing whole extents.
  */
-#define FILE_EXTENT	(256 * 1024)
-#define HLINK_EXTENT	(128 * 1024)
+#define NORMAL_EXTENT	(256 * 1024)
+#define SMALL_EXTENT	(128 * 1024)
 
-#define WITH_HLINK	1
-#define WITHOUT_HLINK	0
+#define FLIST_TEMP	(1<<1)
 
 struct file_list {
-	struct file_struct **files;
+	struct file_list *next, *prev;
+	struct file_struct **files, **sorted;
 	alloc_pool_t file_pool;
-	alloc_pool_t hlink_pool;
-	int count;
-	int malloced;
-	int low, high;
+	void *pool_boundary;
+	int used, malloced;
+	int low, high;  /* 0-relative index values excluding empties */
+	int ndx_start;  /* the start offset for inc_recurse mode */
+	int flist_num;  /* 1-relative file_list number or 0 */
+	int parent_ndx; /* dir_flist index of parent directory */
+	int in_progress, to_redo;
 };
 
 #define SUMFLG_SAME_OFFSET	(1<<0)
@@ -615,10 +781,12 @@ struct map_struct {
 #define MATCHFLG_SENDER_SIDE	(1<<16)/* rule applies to the sending side */
 #define MATCHFLG_RECEIVER_SIDE	(1<<17)/* rule applies to the receiving side */
 #define MATCHFLG_CLEAR_LIST 	(1<<18)/* this item is the "!" token */
+#define MATCHFLG_PERISHABLE	(1<<19)/* perishable if parent dir goes away */
 
 #define MATCHFLGS_FROM_CONTAINER (MATCHFLG_ABS_PATH | MATCHFLG_INCLUDE \
 				| MATCHFLG_DIRECTORY | MATCHFLG_SENDER_SIDE \
-				| MATCHFLG_NEGATE | MATCHFLG_RECEIVER_SIDE)
+				| MATCHFLG_NEGATE | MATCHFLG_RECEIVER_SIDE \
+				| MATCHFLG_PERISHABLE)
 
 struct filter_struct {
 	struct filter_struct *next;
@@ -645,32 +813,92 @@ struct stats {
 	int64 matched_data;
 	int64 flist_buildtime;
 	int64 flist_xfertime;
-	int flist_size;
+	int64 flist_size;
 	int num_files;
 	int num_transferred_files;
-	int current_file_index;
 };
 
 struct chmod_mode_struct;
 
+#define EMPTY_ITEM_LIST {NULL, 0, 0}
+
+typedef struct {
+	void *items;
+	size_t count;
+	size_t malloced;
+} item_list;
+
+#define EXPAND_ITEM_LIST(lp, type, incr) \
+	(type*)expand_item_list(lp, sizeof (type), #type, incr)
+
+#define EMPTY_XBUF {NULL, 0, 0, 0}
+
+typedef struct {
+	char *buf;
+	size_t pos;  /* pos = read pos in the buf */
+	size_t len;  /* len = chars following pos */
+	size_t size; /* size = total space in buf */
+} xbuf;
+
+#define INIT_XBUF(xb, str, ln, sz) (xb).buf = (str), (xb).len = (ln), (xb).size = (sz), (xb).pos = 0
+#define INIT_XBUF_STRLEN(xb, str) (xb).buf = (str), (xb).len = strlen((xb).buf), (xb).size = (-1), (xb).pos = 0
+/* This one is used to make an output xbuf based on a char[] buffer: */
+#define INIT_CONST_XBUF(xb, bf) (xb).buf = (bf), (xb).size = sizeof (bf), (xb).len = (xb).pos = 0
+
+#define ICB_EXPAND_OUT (1<<0)
+#define ICB_INCLUDE_BAD (1<<1)
+#define ICB_INCLUDE_INCOMPLETE (1<<2)
+
+#define RL_EOL_NULLS (1<<0)
+#define RL_DUMP_COMMENTS (1<<1)
+#define RL_CONVERT (1<<2)
+
+typedef struct {
+	char name_type;
+	char fname[1]; /* has variable size */
+} relnamecache;
+
 #include "byteorder.h"
-#include "lib/mdfour.h"
+#include "lib/mdigest.h"
 #include "lib/wildmatch.h"
 #include "lib/permstring.h"
 #include "lib/addrinfo.h"
 
-#if !defined __GNUC__ || defined __APPLE__
-/* Apparently the OS X port of gcc gags on __attribute__.
- *
- * <http://www.opensource.apple.com/bugs/X/gcc/2512150.html> */
+#ifndef __GNUC__
 #define __attribute__(x)
+#else
+# if __GNUC__ <= 2
+# define NORETURN
+# endif
 #endif
 
 #define UNUSED(x) x __attribute__((__unused__))
+#ifndef NORETURN
 #define NORETURN __attribute__((__noreturn__))
+#endif
+
+typedef struct {
+    STRUCT_STAT st;
+#ifdef SUPPORT_ACLS
+    struct rsync_acl *acc_acl; /* access ACL */
+    struct rsync_acl *def_acl; /* default ACL */
+#endif
+#ifdef SUPPORT_XATTRS
+    item_list *xattr;
+#endif
+} stat_x;
+
+#define ACL_READY(sx) ((sx).acc_acl != NULL)
+#define XATTR_READY(sx) ((sx).xattr != NULL)
 
 #include "proto.h"
 
+#ifndef SUPPORT_XATTRS
+#define x_stat(fn,fst,xst) do_stat(fn,fst)
+#define x_lstat(fn,fst,xst) do_lstat(fn,fst)
+#define x_fstat(fd,fst,xst) do_fstat(fd,fst)
+#endif
+
 /* We have replacement versions of these if they're missing. */
 #ifndef HAVE_ASPRINTF
 int asprintf(char **ptr, const char *format, ...);
@@ -687,7 +915,7 @@ int vsnprintf(char *str, size_t count, const char *fmt, va_list args);
 
 #if !defined HAVE_SNPRINTF || !defined HAVE_C99_VSNPRINTF
 #define snprintf rsync_snprintf
-int snprintf(char *str,size_t count,const char *fmt,...);
+int snprintf(char *str, size_t count, const char *fmt,...);
 #endif
 
 
@@ -826,9 +1054,11 @@ extern int errno;
 #define NS(s) ((s)?(s):"<NULL>")
 
 /* Convenient wrappers for malloc and realloc.  Use them. */
-#define new(type) ((type *)malloc(sizeof(type)))
-#define new_array(type, num) ((type *)_new_array(sizeof(type), (num)))
-#define realloc_array(ptr, type, num) ((type *)_realloc_array((ptr), sizeof(type), (num)))
+#define new(type) ((type*)malloc(sizeof (type)))
+#define new0(type) ((type*)calloc(1, sizeof (type)))
+#define new_array(type, num) ((type*)_new_array((num), sizeof (type), 0))
+#define new_array0(type, num) ((type*)_new_array((num), sizeof (type), 1))
+#define realloc_array(ptr, type, num) ((type*)_realloc_array((ptr), sizeof(type), (num)))
 
 /* use magic gcc attributes to catch format errors */
  void rprintf(enum logcode , const char *, ...)
