@@ -3,11 +3,11 @@
  * functions, so that module test harnesses can run standalone.
  *
  * Copyright (C) 2001, 2002 Martin Pool <mbp@samba.org>
- * Copyright (C) 2003, 2004, 2005, 2006 Wayne Davison
+ * Copyright (C) 2003-2008 Wayne Davison
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
@@ -16,8 +16,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
+ * with this program; if not, visit the http://fsf.org website.
  */
 
 #include "rsync.h"
@@ -29,6 +28,7 @@ int human_readable = 0;
 int module_dirlen = 0;
 mode_t orig_umask = 002;
 char *partial_dir;
+char *module_dir;
 struct filter_list_struct server_filter_list;
 
  void rprintf(UNUSED(enum logcode code), const char *format, ...)
@@ -64,6 +64,11 @@ struct filter_list_struct server_filter_list;
 	return 0;
 }
 
+ int make_bak_dir(UNUSED(const char *fullpath))
+{
+	return -1;
+}
+
  char *lp_name(UNUSED(int mod))
 {
 	return NULL;
@@ -74,11 +79,6 @@ struct filter_list_struct server_filter_list;
 	return 0;
 }
 
- char *lp_path(UNUSED(int mod))
-{
-	return NULL;
-}
-
  const char *who_am_i(void)
 {
 	return "tester";
